# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import re
import os
import textwrap
import time
import enum

import psutil

import numpy as np

try:
    import numpy.typing as npt
except ImportError:
    npt = None

import pytest

import tango.asyncio
import tango.constants
from tango import (
    AttrData,
    Attr,
    AttrDataFormat,
    AttrQuality,
    AttReqType,
    AttrWriteType,
    Attribute,
    CmdArgType,
    DevBoolean,
    DevLong,
    DevDouble,
    DevFailed,
    DevEncoded,
    DevEnum,
    DevState,
    DevVoid,
    Device_4Impl,
    Device_5Impl,
    Device_6Impl,
    DeviceClass,
    ExtractAs,
    GreenMode,
    LatestDeviceImpl,
    MultiClassAttribute,
    READ_WRITE,
    SCALAR,
    SPECTRUM,
    EncodedAttribute,
    PyTangoUserWarning,  # noqa
    Util,
    WAttribute,
)
from tango.green import get_executor
from tango.pyutil import TimedAttrData
from tango.server import Device
from tango.server import command, attribute
from tango.test_utils import (
    DeviceTestContext,
    GoodEnum,
    general_decorator,
    general_asyncio_decorator,  # noqa
    assert_close,
    check_attr_type,
    check_read_attr,
    make_nd_value,
    convert_dtype_to_typing_hint,
    UTF8_STRING,
)
from tango.utils import (
    FROM_TANGO_TO_NUMPY_TYPE,
    TO_TANGO_TYPE,
    get_enum_labels,
    is_pure_str,
    get_tango_type_format,
)


@pytest.mark.parametrize("return_time_quality", [True, False])
def test_read_write_attribute_all_types(attribute_typed_values, return_time_quality):
    dtype, values, expected = attribute_typed_values

    class TestDevice(Device):
        _is_allowed = None

        @attribute(
            dtype=dtype, max_dim_x=3, max_dim_y=3, access=AttrWriteType.READ_WRITE
        )
        def attr(self):
            if return_time_quality:
                return self.attr_value, time.time(), AttrQuality.ATTR_VALID
            else:
                return self.attr_value

        @attr.write
        def attr(self, value):
            self.attr_value = value

    with DeviceTestContext(TestDevice) as proxy:
        for value in values:
            proxy.attr = value
            assert_close(proxy.attr, expected(value))


def test_read_write_attribute_with_green_modes(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode
            _is_allowed = None
            attr_value = None

            @attribute(dtype=int, access=AttrWriteType.READ_WRITE)
            async def attr(self):
                return self.attr_value

            @attr.write
            async def attr(self, value):
                self.attr_value = value

            async def is_attr_allowed(self, req_type):
                assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
                return self._is_allowed

            @command(dtype_in=bool)
            async def make_allowed(self, yesno):
                self._is_allowed = yesno

    else:

        class TestDevice(Device):
            green_mode = server_green_mode
            _is_allowed = None
            attr_value = None

            @attribute(dtype=int, access=AttrWriteType.READ_WRITE)
            def attr(self):
                return self.attr_value

            @attr.write
            def attr(self, value):
                self.attr_value = value

            def is_attr_allowed(self, req_type):
                assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
                return self._is_allowed

            @command(dtype_in=bool)
            def make_allowed(self, yesno):
                self._is_allowed = yesno

    with DeviceTestContext(TestDevice) as proxy:
        proxy.make_allowed(True)
        proxy.attr = 1
        assert proxy.attr == 1

        proxy.make_allowed(False)
        with pytest.raises(DevFailed):
            proxy.attr = 1
        with pytest.raises(DevFailed):
            _ = proxy.attr


def test_wrong_encoding_string():
    class TestDevice(Device):
        @attribute(dtype=str)
        def wrong_string(self):
            return "�"

    with DeviceTestContext(TestDevice) as proxy:
        with pytest.raises(DevFailed, match="UnicodeError"):
            _ = proxy.wrong_string


@pytest.mark.parametrize("return_time_quality", [True, False])
def test_attribute_declared_with_typing(attribute_typed_values, return_time_quality):
    dtype, values, expected = attribute_typed_values
    tuple_hint, list_hint, check_x_dim, check_y_dim = convert_dtype_to_typing_hint(
        dtype
    )

    if return_time_quality:
        tuple_hint = tuple[tuple_hint, float, AttrQuality]
        list_hint = list[list_hint, float, AttrQuality]

    class TestDevice(Device):
        attr_value = None

        hint_with_tuple: tuple_hint = attribute(
            access=AttrWriteType.READ_WRITE, fget="read_attr", fset="write_attr"
        )

        user_size_priority_over_hint: tuple_hint = attribute(
            max_dim_x=5,
            max_dim_y=5,
            access=AttrWriteType.READ_WRITE,
            fget="read_attr",
            fset="write_attr",
        )

        hint_with_list: list_hint = attribute(
            max_dim_x=5,
            max_dim_y=5,
            access=AttrWriteType.READ_WRITE,
            fget="read_attr",
            fset="write_attr",
        )

        def read_attr(self):
            if return_time_quality:
                return self.attr_value, time.time(), AttrQuality.ATTR_VALID
            return self.attr_value

        def write_attr(self, value):
            self.attr_value = value

        @attribute(access=AttrWriteType.READ_WRITE)
        def attribute_tuple_hint(self) -> tuple_hint:
            if return_time_quality:
                return self.attr_value, time.time(), AttrQuality.ATTR_VALID
            return self.attr_value

        @attribute_tuple_hint.write
        def attribute_tuple_hint(self, value):
            self.attr_value = value

        @attribute(access=AttrWriteType.READ_WRITE)
        @general_decorator
        def attribute_with_decorated_read_method(self) -> tuple_hint:
            if return_time_quality:
                return self.attr_value, time.time(), AttrQuality.ATTR_VALID
            return self.attr_value

        @attribute_with_decorated_read_method.write
        def attribute_with_decorated_read_method(self, value):
            self.attr_value = value

        @attribute(access=AttrWriteType.READ_WRITE)
        def attribute_tuple_hint_in_write(self):
            if return_time_quality:
                return self.attr_value, time.time(), AttrQuality.ATTR_VALID
            return self.attr_value

        @attribute_tuple_hint_in_write.write
        def attribute_tuple_hint_in_write(self, value: tuple_hint):
            self.attr_value = value

        @attribute(access=AttrWriteType.READ_WRITE)
        def attribute_hint_in_decorated_write_method(self):
            if return_time_quality:
                return self.attr_value, time.time(), AttrQuality.ATTR_VALID
            return self.attr_value

        @attribute_hint_in_decorated_write_method.write
        @general_decorator
        def attribute_hint_in_decorated_write_method(self, value: tuple_hint):
            self.attr_value = value

        @attribute(access=AttrWriteType.READ_WRITE, max_dim_x=5, max_dim_y=5)
        def attribute_user_size_priority_over_hint(self) -> tuple_hint:
            if return_time_quality:
                return self.attr_value, time.time(), AttrQuality.ATTR_VALID
            return self.attr_value

        @attribute_user_size_priority_over_hint.write
        def attribute_user_size_priority_over_hint(self, value):
            self.attr_value = value

        @attribute(access=AttrWriteType.READ_WRITE, max_dim_x=5, max_dim_y=5)
        def attribute_list_hint(self) -> list_hint:
            if return_time_quality:
                return self.attr_value, time.time(), AttrQuality.ATTR_VALID
            return self.attr_value

        @attribute_list_hint.write
        def attribute_list_hint(self, value: tuple_hint):
            self.attr_value = value

        @command()
        def reset(self):
            self.attr_value = None

    def check_attribute_with_size(proxy, attr, value, size_x, size_y):
        setattr(proxy, attr, value)
        assert_close(getattr(proxy, attr), expected(value))
        conf = proxy.get_attribute_config(attr)
        if check_x_dim:
            assert conf.max_dim_x == size_x
        if check_y_dim:
            assert conf.max_dim_y == size_y
        proxy.reset()

    with DeviceTestContext(TestDevice) as proxy:
        for value in values:
            check_attribute_with_size(proxy, "hint_with_tuple", value, 3, 4)
            check_attribute_with_size(
                proxy, "user_size_priority_over_hint", value, 5, 5
            )
            check_attribute_with_size(proxy, "hint_with_list", value, 5, 5)
            check_attribute_with_size(proxy, "attribute_tuple_hint", value, 3, 4)
            check_attribute_with_size(
                proxy, "attribute_with_decorated_read_method", value, 3, 4
            )
            check_attribute_with_size(
                proxy, "attribute_tuple_hint_in_write", value, 3, 4
            )
            check_attribute_with_size(
                proxy, "attribute_hint_in_decorated_write_method", value, 3, 4
            )
            check_attribute_with_size(
                proxy, "attribute_user_size_priority_over_hint", value, 5, 5
            )
            check_attribute_with_size(proxy, "attribute_list_hint", value, 5, 5)


def test_attribute_self_typed_with_not_defined_name():
    _value = [None]

    def non_bound_read(device: "TestDevice") -> int:
        return _value[0]

    def non_bound_read_no_return_hint(device: "TestDevice"):
        return _value[0]

    def non_bound_write(device: "TestDevice", val_in: int):
        _value[0] = val_in

    class TestDevice(Device):
        _value = None

        assignment_attr: int = attribute(
            access=AttrWriteType.READ_WRITE, fget="read_attr", fset="write_attr"
        )

        non_bound_attr: int = attribute(
            access=AttrWriteType.READ_WRITE, fget=non_bound_read, fset=non_bound_write
        )

        non_bound_attr_in_write: int = attribute(
            access=AttrWriteType.READ_WRITE,
            fget=non_bound_read_no_return_hint,
            fset=non_bound_write,
        )

        def read_attr(self: "TestDevice"):
            return self._value

        def write_attr(self: "TestDevice", val_in):
            self._value = val_in

        @attribute
        def decorator_attr(self: "TestDevice") -> int:
            return self._value

        @decorator_attr.write
        def set_value(self: "TestDevice", val_in: int):
            self._value = val_in

        @attribute
        def decorator_attr_def_in_write(self: "TestDevice"):
            return self._value

        @decorator_attr_def_in_write.write
        def set_value_2(self: "TestDevice", val_in: int):
            self._value = val_in

    with DeviceTestContext(TestDevice) as proxy:
        proxy.assignment_attr = 1
        assert 1 == proxy.assignment_attr
        proxy.decorator_attr = 2
        assert 2 == proxy.decorator_attr
        proxy.decorator_attr_def_in_write = 3
        assert 3 == proxy.decorator_attr_def_in_write

        proxy.non_bound_attr = 1
        assert 1 == proxy.non_bound_attr
        proxy.non_bound_attr_in_write = 2
        assert 2 == proxy.non_bound_attr_in_write


def test_read_write_attribute_with_unbound_functions():
    v = {"attr": None}
    is_allowed = None

    def read_attr(device):
        assert isinstance(device, TestDevice)
        return v["attr"]

    def write_attr(device, val):
        assert isinstance(device, TestDevice)
        v["attr"] = val

    def is_attr_allowed(device, req_type):
        assert isinstance(device, TestDevice)
        assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
        return is_allowed

    class TestDevice(Device):

        attr = attribute(
            fget=read_attr,
            fset=write_attr,
            fisallowed=is_attr_allowed,
            dtype=int,
            access=AttrWriteType.READ_WRITE,
        )

    with DeviceTestContext(TestDevice) as proxy:
        is_allowed = True
        proxy.attr = 123
        assert proxy.attr == 123

        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.attr = 123
        with pytest.raises(DevFailed):
            _ = proxy.attr


def test_read_write_attribute_decorated_methods(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class BaseTestDevice(Device):
            @command(dtype_in=bool)
            async def make_allowed(self, yesno):
                self.is_allowed = yesno

    else:

        class BaseTestDevice(Device):
            @command(dtype_in=bool)
            def make_allowed(self, yesno):
                self.is_allowed = yesno

    class TestDevice(BaseTestDevice):
        green_mode = server_green_mode

        attr_value = None
        is_allowed = None

        attr = attribute(dtype=int, access=AttrWriteType.READ_WRITE)

        sync_code = textwrap.dedent(
            """
        @general_decorator
        def read_attr(self):
            return self.attr_value

        @general_decorator
        def write_attr(self, value):
            self.attr_value = value

        @general_decorator
        def is_attr_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self.is_allowed
        """
        )

        if server_green_mode == GreenMode.Asyncio:
            exec(
                sync_code.replace("def", "async def").replace(
                    "general_decorator", "general_asyncio_decorator"
                )
            )
        else:
            exec(sync_code)

    with DeviceTestContext(TestDevice) as proxy:
        proxy.make_allowed(True)
        proxy.attr = 123
        assert proxy.attr == 123

        proxy.make_allowed(False)
        with pytest.raises(DevFailed):
            proxy.attr = 123
        with pytest.raises(DevFailed):
            _ = proxy.attr


def test_read_write_wvalue_attribute(attribute_typed_values):
    dtype, values, expected = attribute_typed_values

    class TestDevice(Device):
        value = None

        attr = attribute(
            dtype=dtype, max_dim_x=3, max_dim_y=3, access=AttrWriteType.READ_WRITE
        )

        def read_attr(self):
            return self.value

        def write_attr(self, value):
            self.value = value
            w_attr = self.get_device_attr().get_w_attr_by_name("attr")
            w_attr.set_write_value(value)

    with DeviceTestContext(TestDevice) as proxy:
        for value in values:
            proxy.attr = value
            assert_close(proxy.attr, expected(proxy.read_attribute("attr").w_value))


def test_get_set_attribute_value_warning_and_alarm_thresholds():

    class TestDevice(Device):

        @attribute(
            dtype=int,
            min_value=-21,
            min_alarm=-11,
            min_warning=-1,
            max_warning=1,
            max_alarm=11,
            max_value=21,
        )
        def attr(self):
            return 0

        @attr.setter
        def attr(self, value):
            pass

        @command()
        def check_limits(self):
            multi_attr = self.get_device_attr()
            attr = multi_attr.get_attr_by_name("attr")

            assert attr.get_min_alarm() == -11
            assert attr.get_max_alarm() == 11
            attr.set_min_alarm(-12)
            attr.set_max_alarm(12)
            assert attr.get_min_alarm() == -12
            assert attr.get_max_alarm() == 12

            assert attr.get_min_warning() == -1
            assert attr.get_max_warning() == 1
            attr.set_min_warning(-2)
            attr.set_max_warning(2)
            assert attr.get_min_warning() == -2
            assert attr.get_max_warning() == 2

            w_attr = multi_attr.get_w_attr_by_name("attr")
            assert w_attr.get_min_value() == -21
            assert w_attr.get_max_value() == 21
            w_attr.set_min_value(-22)
            w_attr.set_max_value(22)
            assert w_attr.get_min_value() == -22
            assert w_attr.get_max_value() == 22

    with DeviceTestContext(TestDevice) as proxy:
        proxy.check_limits()


@pytest.mark.parametrize(
    "input_values",
    [[[], []], [np.empty((0)), np.empty((0, 0))], [np.array([]), np.array([])]],
    ids=["list", "np.empty", "np.array"],
)
def test_write_read_empty_spectrum_image_attribute(extract_as, base_type, input_values):
    requested_type, expected_type = extract_as
    spectrum_value, image_value = input_values

    if requested_type == ExtractAs.Numpy and base_type is str:
        expected_type = tuple

    if (
        requested_type in [ExtractAs.ByteArray, ExtractAs.Bytes, ExtractAs.String]
        and base_type is str
    ):
        pytest.xfail(
            "Conversion from (str,) to ByteArray, Bytes and String not supported. May be fixed in future"
        )

    class TestDevice(Device):
        attr_spectrum_value = spectrum_value
        attr_image_value = image_value

        @attribute(dtype=(base_type,), max_dim_x=3, access=AttrWriteType.READ_WRITE)
        def attr_spectrum(self):
            return self.attr_spectrum_value

        @attr_spectrum.write
        def attr_spectrum(self, value):
            self.attr_spectrum_value = value

        @attribute(
            dtype=((base_type,),),
            max_dim_x=3,
            max_dim_y=3,
            access=AttrWriteType.READ_WRITE,
        )
        def attr_image(self):
            return self.attr_image_value

        @attr_image.write
        def attr_image(self, value):
            self.attr_image_value = value

        @command()
        def check_attr_is_empty_list(self):
            if base_type in [int, float, bool]:
                expected_numpy_type = FROM_TANGO_TO_NUMPY_TYPE[TO_TANGO_TYPE[base_type]]
                assert self.attr_spectrum_value.dtype == np.dtype(expected_numpy_type)
                assert self.attr_image_value.dtype == np.dtype(expected_numpy_type)
            else:
                assert isinstance(self.attr_spectrum_value, list)
                assert isinstance(self.attr_image_value, list)
            assert len(self.attr_spectrum_value) == 0
            assert len(self.attr_image_value) == 0

    with DeviceTestContext(TestDevice) as proxy:
        # first we read init value
        attr_read = proxy.read_attribute("attr_spectrum", extract_as=requested_type)
        assert isinstance(attr_read.value, expected_type)
        assert len(attr_read.value) == 0
        attr_read = proxy.read_attribute("attr_image", extract_as=requested_type)
        assert isinstance(attr_read.value, expected_type)
        assert len(attr_read.value) == 0
        # then we write empty list and check if it was really written
        proxy.attr_spectrum = spectrum_value
        proxy.attr_image = image_value
        proxy.check_attr_is_empty_list()
        # and finally, we read it again and check the value and wvalue
        attr_read = proxy.read_attribute("attr_spectrum", extract_as=requested_type)
        assert isinstance(attr_read.value, expected_type)
        assert len(attr_read.value) == 0
        assert isinstance(attr_read.w_value, expected_type)
        assert len(attr_read.w_value) == 0
        attr_read = proxy.read_attribute("attr_image", extract_as=requested_type)
        assert isinstance(attr_read.value, expected_type)
        assert len(attr_read.value) == 0
        assert isinstance(attr_read.w_value, expected_type)
        assert len(attr_read.w_value) == 0


@pytest.mark.parametrize(
    "device_impl_class", [Device_4Impl, Device_5Impl, Device_6Impl, LatestDeviceImpl]
)
def test_write_read_empty_spectrum_attribute_classic_api(
    device_impl_class, extract_as, base_type
):
    requested_type, expected_type = extract_as

    if requested_type == ExtractAs.Numpy and base_type is str:
        expected_type = tuple

    if (
        requested_type in [ExtractAs.ByteArray, ExtractAs.Bytes, ExtractAs.String]
        and base_type is str
    ):
        pytest.xfail(
            "Conversion from (str,) to ByteArray, Bytes and String not supported. May be fixed in future"
        )

    class ClassicAPIClass(DeviceClass):
        cmd_list = {"check_attr_is_empty_list": [[DevVoid, "none"], [DevVoid, "none"]]}
        attr_list = {
            "attr": [[TO_TANGO_TYPE[base_type], SPECTRUM, AttrWriteType.READ_WRITE, 10]]
        }

        def __init__(self, name):
            super().__init__(name)
            self.set_type("TestDevice")

    class ClassicAPIDeviceImpl(device_impl_class):
        attr_value = []

        def read_attr(self, attr):
            attr.set_value(self.attr_value)

        def write_attr(self, attr):
            w_value = attr.get_write_value()
            self.attr_value = w_value

        def check_attr_is_empty_list(self):
            if base_type in [int, float, bool]:
                expected_numpy_type = FROM_TANGO_TO_NUMPY_TYPE[TO_TANGO_TYPE[base_type]]
                assert self.attr_value.dtype == np.dtype(expected_numpy_type)
            else:
                assert isinstance(self.attr_value, list)
            assert len(self.attr_value) == 0

    with DeviceTestContext(ClassicAPIDeviceImpl, ClassicAPIClass) as proxy:
        # first we read init value
        attr_read = proxy.read_attribute("attr", extract_as=requested_type)
        assert isinstance(attr_read.value, expected_type)
        assert len(attr_read.value) == 0
        # then we write empty list and check if it was really written
        proxy.attr = []
        proxy.check_attr_is_empty_list()
        # and finally, we read it again and check the value and wvalue
        attr_read = proxy.read_attribute("attr", extract_as=requested_type)
        assert isinstance(attr_read.value, expected_type)
        assert len(attr_read.value) == 0
        assert isinstance(attr_read.w_value, expected_type)
        assert len(attr_read.w_value) == 0


@pytest.mark.parametrize("dtype", ["state", DevState, CmdArgType.DevState])
def test_ensure_devstate_is_pytango_enum(attr_data_format, dtype):
    if attr_data_format == AttrDataFormat.SCALAR:
        value = DevState.ON
    elif attr_data_format == AttrDataFormat.SPECTRUM:
        dtype = (dtype,)
        value = (DevState.ON, DevState.RUNNING)
    else:
        dtype = ((dtype,),)
        value = ((DevState.ON, DevState.RUNNING), (DevState.UNKNOWN, DevState.MOVING))

    class TestDevice(Device):
        @attribute(dtype=dtype, access=AttrWriteType.READ, max_dim_x=3, max_dim_y=3)
        def any_name_for_state_attribute(self):
            return value

    with DeviceTestContext(TestDevice) as proxy:
        states = proxy.any_name_for_state_attribute
        assert states == value
        if attr_data_format == AttrDataFormat.SCALAR:
            assert states is value
        check_attr_type(states, attr_data_format, DevState)


def test_read_write_attribute_enum(attr_data_format):
    values = [member.value for member in GoodEnum]
    enum_labels = get_enum_labels(GoodEnum)

    if attr_data_format == AttrDataFormat.SCALAR:
        good_type = GoodEnum
        good_type_str = "DevEnum"
    elif attr_data_format == AttrDataFormat.SPECTRUM:
        good_type = (GoodEnum,)
        good_type_str = ("DevEnum",)
    else:
        good_type = ((GoodEnum,),)
        good_type_str = (("DevEnum",),)

    class TestDevice(Device):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            if attr_data_format == AttrDataFormat.SCALAR:
                self.attr_from_enum_value = 0
                self.attr_from_labels_value = 0
            elif attr_data_format == AttrDataFormat.SPECTRUM:
                self.attr_from_enum_value = (0,)
                self.attr_from_labels_value = (0,)
            else:
                self.attr_from_enum_value = ((0,),)
                self.attr_from_labels_value = ((0,),)

        attr_from_enum = attribute(
            dtype=good_type, max_dim_x=3, max_dim_y=3, access=AttrWriteType.READ_WRITE
        )

        attr_from_labels = attribute(
            dtype=good_type_str,
            max_dim_x=3,
            max_dim_y=3,
            enum_labels=enum_labels,
            access=AttrWriteType.READ_WRITE,
        )

        def read_attr_from_enum(self):
            return self.attr_from_enum_value

        def write_attr_from_enum(self, value):
            self.attr_from_enum_value = value

        def read_attr_from_labels(self):
            return self.attr_from_labels_value

        def write_attr_from_labels(self, value):
            self.attr_from_labels_value = value

    with DeviceTestContext(TestDevice) as proxy:
        # test assigning values (ints)
        for value, label in zip(values, enum_labels):
            nd_value = make_nd_value(value, attr_data_format)
            proxy.attr_from_enum = nd_value
            read_attr = proxy.attr_from_enum
            assert read_attr == nd_value
            check_attr_type(read_attr, attr_data_format, enum.IntEnum)
            check_read_attr(read_attr, attr_data_format, value, label)

            proxy.attr_from_labels = nd_value
            read_attr = proxy.attr_from_labels
            assert read_attr == nd_value
            check_attr_type(read_attr, attr_data_format, enum.IntEnum)
            check_read_attr(read_attr, attr_data_format, value, label)

        # test assigning labels (strings)
        for value, label in zip(values, enum_labels):
            nd_label = make_nd_value(label, attr_data_format)
            proxy.attr_from_enum = nd_label
            read_attr = proxy.attr_from_enum
            check_attr_type(read_attr, attr_data_format, enum.IntEnum)
            check_read_attr(read_attr, attr_data_format, value, label)

            proxy.attr_from_labels = nd_label
            read_attr = proxy.attr_from_labels
            check_attr_type(read_attr, attr_data_format, enum.IntEnum)
            check_read_attr(read_attr, attr_data_format, value, label)

        invalid_label = make_nd_value("_DOES_NOT_EXIST_", attr_data_format)
        expected_match = re.escape(f"Valid values: {enum_labels}")
        with pytest.raises(AttributeError, match=expected_match):
            proxy.attr_from_enum = invalid_label
        with pytest.raises(AttributeError, match=expected_match):
            proxy.attr_from_labels = invalid_label

    with pytest.raises(TypeError) as context:

        class BadTestDevice(Device):

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                if attr_data_format == AttrDataFormat.SCALAR:
                    self.attr_value = 0
                elif attr_data_format == AttrDataFormat.SPECTRUM:
                    self.attr_value = (0,)
                else:
                    self.attr_value = ((0,),)

            # enum_labels may not be specified if dtype is an enum.Enum
            @attribute(
                dtype=good_type, max_dim_x=3, max_dim_y=3, enum_labels=enum_labels
            )
            def bad_attr(self):
                return self.attr_value

        BadTestDevice()  # dummy instance for Codacy
    assert "enum_labels" in str(context.value)


@pytest.mark.parametrize("enum_type", [DevState, GoodEnum])
def test_enum_devstate_attribute_declared_with_typing(attr_data_format, enum_type):
    value = DevState.MOVING if enum_type is DevState else GoodEnum.MIDDLE
    expected_type = DevState if enum_type is DevState else enum.IntEnum
    nd_value = make_nd_value(value, attr_data_format)

    if attr_data_format == AttrDataFormat.SCALAR:
        EnumType = enum_type
    elif attr_data_format == AttrDataFormat.SPECTRUM:
        EnumType = tuple[enum_type, enum_type, enum_type]
    else:
        EnumType = tuple[
            tuple[enum_type, enum_type, enum_type],
            tuple[enum_type, enum_type, enum_type],
            tuple[enum_type, enum_type, enum_type],
        ]

    class TestDevice(Device):
        attr: EnumType = attribute(access=AttrWriteType.READ)

        def read_attr(self):
            return nd_value

    with DeviceTestContext(TestDevice) as proxy:
        read_value = proxy.attr
        assert read_value == nd_value
        check_attr_type(read_value, attr_data_format, expected_type)
        if enum_type is GoodEnum:
            check_read_attr(read_value, attr_data_format, value, "MIDDLE")


def test_read_attribute_with_invalid_quality_is_none(attribute_typed_values):
    dtype, values, expected = attribute_typed_values

    class TestDevice(Device):
        @attribute(dtype=dtype, max_dim_x=3, max_dim_y=3)
        def attr(self):
            dummy_time = 123.4
            return values[0], dummy_time, AttrQuality.ATTR_INVALID

    with DeviceTestContext(TestDevice) as proxy:
        reading = proxy.read_attribute("attr")
        assert reading.value is None
        assert reading.quality == AttrQuality.ATTR_INVALID
        high_level_value = proxy.attr
        assert high_level_value is None


def test_read_enum_attribute_with_invalid_quality_is_none():
    class TestDevice(Device):
        @attribute(dtype=GoodEnum)
        def attr(self):
            dummy_time = 123.4
            return GoodEnum.START, dummy_time, AttrQuality.ATTR_INVALID

    with DeviceTestContext(TestDevice) as proxy:
        reading = proxy.read_attribute("attr")
        assert reading.value is None
        assert reading.quality == AttrQuality.ATTR_INVALID
        high_level_value = proxy.attr
        assert high_level_value is None


def test_wrong_attribute_read():
    class TestDevice(Device):

        @attribute(dtype=str)
        def attr_str_err(self):
            return 1.2345

        @attribute(dtype=int)
        def attr_int_err(self):
            return "bla"

        @attribute(dtype=[str])
        def attr_str_list_err(self):
            return ["hello", 55]

    with DeviceTestContext(TestDevice) as proxy:
        with pytest.raises(DevFailed):
            proxy.attr_str_err
        with pytest.raises(DevFailed):
            proxy.attr_int_err
        with pytest.raises(DevFailed):
            proxy.attr_str_list_err


def test_attribute_access_with_default_method_names():

    is_allowed = True

    class TestDevice(Device):

        _read_write_value = ""
        _is_allowed = True

        attr_r = attribute(dtype=str)
        attr_rw = attribute(dtype=str, access=AttrWriteType.READ_WRITE)

        def read_attr_r(self):
            return "readable"

        def read_attr_rw(self):
            print(f"Return value {self._read_write_value}")
            return self._read_write_value

        def write_attr_rw(self, value):
            print(f"Get value {value}")
            self._read_write_value = value

        def is_attr_r_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return is_allowed

        def is_attr_rw_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return is_allowed

    with DeviceTestContext(TestDevice) as proxy:
        with pytest.raises(DevFailed):
            proxy.attr_r = "writable"
        assert proxy.attr_r == "readable"
        proxy.attr_rw = "writable"
        assert proxy.attr_rw == "writable"

        is_allowed = False
        with pytest.raises(DevFailed):
            _ = proxy.attr_r
        with pytest.raises(DevFailed):
            proxy.attr_rw = "writing_not_allowed"
        with pytest.raises(DevFailed):
            _ = proxy.attr_rw


@pytest.fixture(
    ids=["low_level_read", "high_level_read"],
    params=[
        textwrap.dedent(
            """\
                        def read_dyn_attr(self, attr):
                            attr.set_value(self.attr_value)
                            """
        ),
        textwrap.dedent(
            """\
                        def read_dyn_attr(self, attr):
                            return self.attr_value
                            """
        ),
    ],
)
def dynamic_attribute_read_function(request):
    return request.param


def test_read_write_dynamic_attribute(
    dynamic_attribute_read_function, server_green_mode
):
    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode
            attr_value = None

            @command
            async def add_dyn_attr(self):
                attr = attribute(
                    name="dyn_attr",
                    dtype=int,
                    access=AttrWriteType.READ_WRITE,
                    fget=self.read_dyn_attr,
                    fset=self.write_dyn_attr,
                )
                await self.async_add_attribute(attr)

            @command
            async def delete_dyn_attr(self):
                await self.async_remove_attribute("dyn_attr")

            async def write_dyn_attr(self, attr):
                self.attr_value = attr.get_write_value()

            exec(dynamic_attribute_read_function.replace("def ", "async def "))

    else:

        class TestDevice(Device):
            green_mode = server_green_mode
            attr_value = None

            @command
            def add_dyn_attr(self):
                attr = attribute(
                    name="dyn_attr",
                    dtype=int,
                    access=AttrWriteType.READ_WRITE,
                    fget=self.read_dyn_attr,
                    fset=self.write_dyn_attr,
                )
                self.add_attribute(attr)

            @command
            def delete_dyn_attr(self):
                self.remove_attribute("dyn_attr")

            def write_dyn_attr(self, attr):
                self.attr_value = attr.get_write_value()

            exec(dynamic_attribute_read_function)

    with DeviceTestContext(TestDevice) as proxy:
        proxy.add_dyn_attr()
        proxy.dyn_attr = 123
        assert proxy.dyn_attr == 123
        proxy.delete_dyn_attr()
        assert "dyn_attr" not in proxy.get_attribute_list()


def test_async_add_remove_dynamic_attribute():
    class TestDevice(Device):
        green_mode = GreenMode.Asyncio

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.attr_value = None

        @command
        async def add_dyn_attr(self):
            attr = attribute(
                name="dyn_attr",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget=self.read_dyn_attr,
                fset=self.write_dyn_attr,
            )
            self.add_attribute(attr)

        @command
        async def delete_dyn_attr(self):
            self.remove_attribute("dyn_attr")

        @command
        async def async_add_dyn_attr(self):
            attr = attribute(
                name="dyn_attr",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget=self.read_dyn_attr,
                fset=self.write_dyn_attr,
            )
            await self.async_add_attribute(attr)

        @command
        async def async_delete_dyn_attr(self):
            await self.async_remove_attribute("dyn_attr")

        async def write_dyn_attr(self, attr):
            self.attr_value = attr.get_write_value()

        async def read_dyn_attr(self, attr):
            return self.attr_value

    with DeviceTestContext(TestDevice) as proxy:
        proxy.add_dyn_attr()
        proxy.dyn_attr = 123
        assert proxy.dyn_attr == 123
        proxy.delete_dyn_attr()
        assert "dyn_attr" not in proxy.get_attribute_list()

        proxy.async_add_dyn_attr()
        proxy.dyn_attr = 123
        assert proxy.dyn_attr == 123
        proxy.async_delete_dyn_attr()
        assert "dyn_attr" not in proxy.get_attribute_list()


def test_dynamic_attribute_declared_with_typing(attribute_typed_values):
    dtype, values, expected = attribute_typed_values
    tuple_hint, list_hint, check_x_dim, check_y_dim = convert_dtype_to_typing_hint(
        dtype
    )

    class TestDevice(Device):
        attr_value = None

        def initialize_dynamic_attributes(self):
            attr = attribute(
                name="read_function_tuple_hint",
                access=AttrWriteType.READ_WRITE,
                fget=self.read_attr_with_tuple_hints,
                fset=self.write_attr_no_hints,
            )
            self.add_attribute(attr)

            attr = attribute(
                name="user_size_priority_over_hint",
                max_dim_x=5,
                max_dim_y=5,
                access=AttrWriteType.READ_WRITE,
                fget=self.read_attr_with_tuple_hints,
                fset=self.write_attr_no_hints,
            )
            self.add_attribute(attr)

            attr = attribute(
                name="read_function_list_hint",
                access=AttrWriteType.READ_WRITE,
                max_dim_x=5,
                max_dim_y=5,
                fget=self.read_attr_with_list_hints,
                fset=self.write_attr_no_hints,
            )
            self.add_attribute(attr)

            attr = attribute(
                name="write_function_tuple_hint",
                access=AttrWriteType.READ_WRITE,
                fget=self.read_attr_no_hints,
                fset=self.write_attr_with_tuple_hints,
            )
            self.add_attribute(attr)

            attr = attribute(
                name="write_function_list_hint",
                access=AttrWriteType.READ_WRITE,
                max_dim_x=5,
                max_dim_y=5,
                fget=self.read_attr_no_hints,
                fset=self.write_attr_with_list_hints,
            )
            self.add_attribute(attr)

        def read_attr_no_hints(self, attr):
            return self.attr_value

        def write_attr_no_hints(self, attr):
            self.attr_value = attr.get_write_value()

        def read_attr_with_tuple_hints(self, attr) -> tuple_hint:
            return self.attr_value

        def read_attr_with_list_hints(self, attr) -> list_hint:
            return self.attr_value

        def write_attr_with_tuple_hints(self, attr: tuple_hint):
            self.attr_value = attr.get_write_value()

        def write_attr_with_list_hints(self, attr: list_hint):
            self.attr_value = attr.get_write_value()

        @command()
        def reset(self):
            self.attr_value = None

    def check_attribute_with_size(proxy, attr, value, size_x, size_y):
        setattr(proxy, attr, value)
        assert_close(getattr(proxy, attr), expected(value))
        conf = proxy.get_attribute_config(attr)
        if check_x_dim:
            assert conf.max_dim_x == size_x
        if check_y_dim:
            assert conf.max_dim_y == size_y
        proxy.reset()

    with DeviceTestContext(TestDevice) as proxy:
        for value in values:
            check_attribute_with_size(proxy, "read_function_tuple_hint", value, 3, 4)
            check_attribute_with_size(proxy, "read_function_list_hint", value, 5, 5)
            check_attribute_with_size(
                proxy, "user_size_priority_over_hint", value, 5, 5
            )
            check_attribute_with_size(proxy, "write_function_tuple_hint", value, 3, 4)
            check_attribute_with_size(proxy, "write_function_list_hint", value, 5, 5)


def test_dynamic_attribute_self_typed_with_not_defined_name():
    _value = [None]

    def non_bound_read(device: "TestDevice", attr) -> int:
        return _value[0]

    def non_bound_read_no_return_hint(device: "TestDevice", attr):
        return _value[0]

    def non_bound_write(device: "TestDevice", attr: int):
        _value[0] = attr.get_write_value()

    class TestDevice(Device):
        _value = None

        def initialize_dynamic_attributes(self):
            attr = attribute(
                name="read_with_hint",
                access=AttrWriteType.READ_WRITE,
                fget=self.read_attr,
                fset=self.write_attr,
            )
            self.add_attribute(attr)

            attr = attribute(
                name="read_no_hint",
                access=AttrWriteType.READ_WRITE,
                fget=self.read_attr_no_hint,
                fset=self.write_attr,
            )
            self.add_attribute(attr)

            attr = attribute(
                name="non_bound_read_with_hint",
                access=AttrWriteType.READ_WRITE,
                fget=non_bound_read,
                fset=non_bound_write,
            )
            self.add_attribute(attr)

            attr = attribute(
                name="non_bound_read_no_hint",
                access=AttrWriteType.READ_WRITE,
                fget=non_bound_read_no_return_hint,
                fset=non_bound_write,
            )
            self.add_attribute(attr)

        def read_attr(self: "TestDevice", attr) -> int:
            return self._value

        def read_attr_no_hint(self: "TestDevice", attr):
            return self._value

        def write_attr(self: "TestDevice", attr: int):
            self._value = attr.get_write_value()

    with DeviceTestContext(TestDevice) as proxy:
        proxy.read_with_hint = 1
        assert 1 == proxy.read_with_hint
        proxy.read_no_hint = 2
        assert 2 == proxy.read_no_hint

        proxy.non_bound_read_with_hint = 1
        assert 1 == proxy.non_bound_read_with_hint
        proxy.non_bound_read_no_hint = 2
        assert 2 == proxy.non_bound_read_no_hint


if npt:

    def test_attribute_declared_with_numpy_typing(attribute_numpy_typed_values):
        type_hint, dformat, value, expected = attribute_numpy_typed_values

        class TestDevice(Device):
            attr_value = None

            statement_declaration: type_hint = attribute(
                access=AttrWriteType.READ_WRITE,
                fget="read_attr",
                fset="write_attr",
                dformat=dformat,
                max_dim_x=2,
                max_dim_y=2,
            )

            def read_attr(self):
                return self.attr_value

            def write_attr(self, value):
                self.attr_value = value

            def initialize_dynamic_attributes(self):
                attr = attribute(
                    name="dynamic_declaration",
                    access=AttrWriteType.READ_WRITE,
                    fget=self.read_dynamic_attr,
                    fset=self.write_dynamic_attr,
                    dformat=dformat,
                    max_dim_x=2,
                    max_dim_y=2,
                )
                self.add_attribute(attr)

            def read_dynamic_attr(self, attr):
                return self.attr_value

            def write_dynamic_attr(self, attr):
                self.attr_value = attr.get_write_value()

            @attribute(
                access=AttrWriteType.READ_WRITE,
                dformat=dformat,
                max_dim_x=2,
                max_dim_y=2,
            )
            def decorator_declaration(self) -> type_hint:
                return self.attr_value

            @decorator_declaration.write
            def decorator_declaration_write(self, value: type_hint):
                self.attr_value = value

            @command()
            def reset(self):
                self.attr_value = None

        def check_attribute(proxy, attr, value):
            setattr(proxy, attr, value)
            assert_close(getattr(proxy, attr), expected(value))
            proxy.reset()

        with DeviceTestContext(TestDevice) as proxy:
            check_attribute(proxy, "statement_declaration", value)
            check_attribute(proxy, "decorator_declaration", value)
            check_attribute(proxy, "dynamic_declaration", value)

    def test_attribute_wrong_declared_with_numpy_typing(attribute_wrong_numpy_typed):
        dformat, max_x, max_y, value, error, match = attribute_wrong_numpy_typed

        with pytest.raises(error, match=match):

            class TestDevice(Device):
                attr_value = None

                attr: npt.NDArray[np.int_] = attribute(
                    access=AttrWriteType.READ_WRITE,
                    dformat=dformat,
                    max_dim_x=max_x,
                    max_dim_y=max_y,
                )

                def read_attr(self):
                    return self.attr_value

                def write_attr(self, value):
                    self.attr_value = value

            with DeviceTestContext(TestDevice) as proxy:
                proxy.attr = value
                _ = proxy.attr


def test_read_write_dynamic_attribute_decorated_methods_default_names(
    server_green_mode,
):

    is_allowed = True

    class TestDevice(Device):
        green_mode = server_green_mode

        attr_value = None
        is_allowed = None

        def initialize_dynamic_attributes(self):
            attr = attribute(name="attr", dtype=int, access=AttrWriteType.READ_WRITE)
            self.add_attribute(attr)

        def allowed(self):
            return is_allowed

        sync_code = textwrap.dedent(
            """\
        @general_decorator
        def read_attr(self, attr):
            return self.attr_value

        @general_decorator
        def write_attr(self, attr):
            self.attr_value = attr.get_write_value()

        @general_decorator
        def is_attr_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self.allowed()
        """
        )

        if server_green_mode != GreenMode.Asyncio:
            exec(sync_code)
        else:
            exec(
                sync_code.replace("def ", "async def ").replace(
                    "general_decorator", "general_asyncio_decorator"
                )
            )

    with DeviceTestContext(TestDevice) as proxy:
        proxy.attr = 123
        assert proxy.attr == 123

        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.attr = 123
        with pytest.raises(DevFailed):
            _ = proxy.attr


def test_read_write_dynamic_attribute_decorated_methods_user_names(server_green_mode):

    is_allowed = True

    class TestDevice(Device):
        green_mode = server_green_mode

        attr_value = None
        is_allowed = None

        def initialize_dynamic_attributes(self):
            attr = attribute(
                name="attr",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget=self.user_read,
                fset=self.user_write,
                fisallowed=self.user_is_allowed,
            )
            self.add_attribute(attr)

        def allowed(self):
            return is_allowed

        sync_code = textwrap.dedent(
            """\
        @general_decorator
        def user_read(self, attr):
            return self.attr_value

        @general_decorator
        def user_write(self, attr):
            self.attr_value = attr.get_write_value()

        @general_decorator
        def user_is_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self.allowed()
        """
        )

        if server_green_mode != GreenMode.Asyncio:
            exec(sync_code)
        else:
            exec(
                sync_code.replace("def ", "async def ").replace(
                    "general_decorator", "general_asyncio_decorator"
                )
            )

    with DeviceTestContext(TestDevice) as proxy:
        proxy.attr = 123
        assert proxy.attr == 123

        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.attr = 123
        with pytest.raises(DevFailed):
            _ = proxy.attr


def test_read_write_dynamic_attribute_decorated_shared_user_functions():

    is_allowed = True

    class TestDevice(Device):

        attr_values = {"attr1": None, "attr2": None}
        is_allowed = None

        def initialize_dynamic_attributes(self):
            attr = attribute(
                name="attr1",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget=self.user_read,
                fset=self.user_write,
                fisallowed=self.user_is_allowed,
            )
            self.add_attribute(attr)
            attr = attribute(
                name="attr2",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget=self.user_read,
                fset=self.user_write,
                fisallowed=self.user_is_allowed,
            )
            self.add_attribute(attr)

        @general_decorator
        def user_read(self, attr):
            return self.attr_values[attr.get_name()]

        @general_decorator
        def user_write(self, attr):
            self.attr_values[attr.get_name()] = attr.get_write_value()

        @general_decorator
        def user_is_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return is_allowed

    with DeviceTestContext(TestDevice) as proxy:
        proxy.attr1 = 123
        assert proxy.attr1 == 123
        proxy.attr2 = 456
        assert proxy.attr1 == 123
        assert proxy.attr2 == 456

        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.attr1 = 123
        with pytest.raises(DevFailed):
            _ = proxy.attr1
        with pytest.raises(DevFailed):
            proxy.attr2 = 123
        with pytest.raises(DevFailed):
            _ = proxy.attr2


def test_read_write_dynamic_attribute_enum(attr_data_format):
    values = [member.value for member in GoodEnum]
    enum_labels = get_enum_labels(GoodEnum)

    if attr_data_format == AttrDataFormat.SCALAR:
        attr_type = DevEnum
        attr_info = (DevEnum, attr_data_format, READ_WRITE)
    elif attr_data_format == AttrDataFormat.SPECTRUM:
        attr_type = (DevEnum,)
        attr_info = (DevEnum, attr_data_format, READ_WRITE, 10)
    else:
        attr_type = ((DevEnum,),)
        attr_info = (DevEnum, attr_data_format, READ_WRITE, 10, 10)

    class TestDevice(Device):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            if attr_data_format == AttrDataFormat.SCALAR:
                self.attr_value = 0
            elif attr_data_format == AttrDataFormat.SPECTRUM:
                self.attr_value = (0,)
            else:
                self.attr_value = ((0,),)

        @command
        def add_dyn_attr_old(self):
            attr = AttrData(
                "dyn_attr",
                None,
                attr_info=[
                    attr_info,
                    {"enum_labels": enum_labels},
                ],
            )
            self.add_attribute(
                attr, r_meth=self.read_dyn_attr, w_meth=self.write_dyn_attr
            )

        @command
        def add_dyn_attr_new(self):
            attr = attribute(
                name="dyn_attr",
                dtype=attr_type,
                enum_labels=enum_labels,
                max_dim_x=3,
                max_dim_y=3,
                access=AttrWriteType.READ_WRITE,
                fget=self.read_dyn_attr,
                fset=self.write_dyn_attr,
            )
            self.add_attribute(attr)

        @command
        def delete_dyn_attr(self):
            self.remove_attribute("dyn_attr")

        def read_dyn_attr(self, attr):
            return self.attr_value

        def write_dyn_attr(self, attr):
            self.attr_value = attr.get_write_value()

    with DeviceTestContext(TestDevice) as proxy:
        for add_attr_cmd in [proxy.add_dyn_attr_old, proxy.add_dyn_attr_new]:
            add_attr_cmd()
            for value, label in zip(values, enum_labels):
                nd_value = make_nd_value(value, attr_data_format)
                proxy.dyn_attr = nd_value
                read_attr = proxy.dyn_attr
                assert read_attr == nd_value
                check_attr_type(read_attr, attr_data_format, enum.IntEnum)
                check_read_attr(read_attr, attr_data_format, value, label)
            proxy.delete_dyn_attr()
            assert "dyn_attr" not in proxy.get_attribute_list()


@pytest.mark.parametrize("enum_type", [DevState, GoodEnum])
def test_enum_devstate_dynamic_attribute_declared_with_typing(
    attr_data_format, enum_type
):
    value = DevState.MOVING if enum_type is DevState else GoodEnum.MIDDLE
    expected_type = DevState if enum_type is DevState else enum.IntEnum
    nd_value = make_nd_value(value, attr_data_format)

    if attr_data_format == AttrDataFormat.SCALAR:
        EnumType = enum_type
    elif attr_data_format == AttrDataFormat.SPECTRUM:
        EnumType = tuple[enum_type, enum_type, enum_type]
    else:
        EnumType = tuple[
            tuple[enum_type, enum_type, enum_type],
            tuple[enum_type, enum_type, enum_type],
        ]

    class TestDevice(Device):
        def initialize_dynamic_attributes(self):
            self.add_attribute(attribute(name="attr", access=AttrWriteType.READ))

        def read_attr(self, attr) -> EnumType:
            return nd_value

    with DeviceTestContext(TestDevice) as proxy:
        read_value = proxy.attr
        assert read_value == nd_value
        check_attr_type(read_value, attr_data_format, expected_type)
        if enum_type is GoodEnum:
            check_read_attr(read_value, attr_data_format, value, "MIDDLE")


def test_read_write_dynamic_attribute_is_allowed_with_async(server_green_mode):
    DYN_ATTRS_END_RANGE = 9

    if server_green_mode == GreenMode.Asyncio:

        class BaseTestDevice(Device):
            @command(dtype_in=bool)
            async def make_allowed(self, yesno):
                for att_num in range(1, DYN_ATTRS_END_RANGE):
                    setattr(self, f"attr{att_num}_allowed", yesno)

    else:

        class BaseTestDevice(Device):
            @command(dtype_in=bool)
            def make_allowed(self, yesno):
                for att_num in range(1, DYN_ATTRS_END_RANGE):
                    setattr(self, f"attr{att_num}_allowed", yesno)

    class TestDevice(BaseTestDevice):
        green_mode = server_green_mode

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            for att_num in range(1, DYN_ATTRS_END_RANGE):
                setattr(self, f"attr{att_num}_allowed", True)
            for att_num in range(1, DYN_ATTRS_END_RANGE):
                setattr(self, f"attr{att_num}_value", None)

        def initialize_dynamic_attributes(self):
            # recommended approach: using attribute() and bound methods:
            attr = attribute(
                name="dyn_attr1",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget=self.read_dyn_attr1,
                fset=self.write_dyn_attr1,
                fisallowed=self.is_attr1_allowed,
            )
            self.add_attribute(attr)

            # not recommended: using attribute() with unbound methods:
            attr = attribute(
                name="dyn_attr2",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget=TestDevice.read_dyn_attr2,
                fset=TestDevice.write_dyn_attr2,
                fisallowed=TestDevice.is_attr2_allowed,
            )
            self.add_attribute(attr)

            # possible approach: using attribute() with method name strings:
            attr = attribute(
                name="dyn_attr3",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget="read_dyn_attr3",
                fset="write_dyn_attr3",
                fisallowed="is_attr3_allowed",
            )
            self.add_attribute(attr)

            # old approach: using tango.AttrData with bound methods:
            attr_name = "dyn_attr4"
            data_info = self._get_attr_data_info()
            dev_class = self.get_device_class()
            attr_data = AttrData(attr_name, dev_class.get_name(), data_info)
            self.add_attribute(
                attr_data,
                self.read_dyn_attr4,
                self.write_dyn_attr4,
                self.is_attr4_allowed,
            )

            # old approach: using tango.AttrData with unbound methods:
            attr_name = "dyn_attr5"
            attr_data = AttrData(attr_name, dev_class.get_name(), data_info)
            self.add_attribute(
                attr_data,
                TestDevice.read_dyn_attr5,
                TestDevice.write_dyn_attr5,
                TestDevice.is_attr5_allowed,
            )

            # old approach: using tango.AttrData with default method names
            attr_name = "dyn_attr6"
            attr_data = AttrData(attr_name, dev_class.get_name(), data_info)
            self.add_attribute(attr_data)

            # old approach: using tango.AttrData filled from dictionary
            attr_name = "dyn_attr7"
            d = {
                "name": attr_name,
                "class_name": dev_class.get_name(),
                # not setting access explicitly
                "fread": "read_dyn_attr7",
                "fwrite": "write_dyn_attr7",
                "fisallowed": self.is_attr7_allowed,
            }
            attr_data = AttrData.from_dict(d)
            self.add_attribute(attr_data)

            # not recommened: implicit access level
            attr = attribute(
                name="dyn_attr8",
                fset=self.write_dyn_attr8,
                fisallowed=self.is_attr8_allowed,
            )
            self.add_attribute(attr)

        def _get_attr_data_info(self):
            simple_type, fmt = get_tango_type_format(int)
            data_info = [[simple_type, fmt, READ_WRITE]]
            return data_info

        # the following methods are written in plain text which looks
        # weird. This is done so that it is easy to change for async
        # tests without duplicating all the code.
        read_code = textwrap.dedent(
            """
        def read_dyn_attr(self, attr):
            return self.attr_value
        """
        )

        write_code = textwrap.dedent(
            """
        def write_dyn_attr(self, attr):
            self.attr_value = attr.get_write_value()
        """
        )

        is_allowed_code = textwrap.dedent(
            """
        def is_attr_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self.attr_allowed
        """
        )

        for attr_num in range(1, DYN_ATTRS_END_RANGE):
            read_method = read_code.replace("read_dyn_attr", f"read_dyn_attr{attr_num}")
            read_method = read_method.replace("attr_value", f"attr{attr_num}_value")
            write_method = write_code.replace(
                "write_dyn_attr", f"write_dyn_attr{attr_num}"
            )
            write_method = write_method.replace("attr_value", f"attr{attr_num}_value")
            if attr_num != 6:
                is_allowed_method = is_allowed_code.replace(
                    "is_attr_allowed", f"is_attr{attr_num}_allowed"
                )
            else:
                # default name differs
                is_allowed_method = is_allowed_code.replace(
                    "is_attr_allowed", f"is_dyn_attr{attr_num}_allowed"
                )
            is_allowed_method = is_allowed_method.replace(
                "self.attr_allowed", f"self.attr{attr_num}_allowed"
            )

            if server_green_mode != GreenMode.Asyncio:
                exec(read_method)
                exec(write_method)
                exec(is_allowed_method)
            else:
                exec(read_method.replace("def ", "async def "))
                exec(write_method.replace("def ", "async def "))
                exec(is_allowed_method.replace("def ", "async def "))

    with DeviceTestContext(TestDevice) as proxy:
        proxy.make_allowed(True)

        for ind in range(1, DYN_ATTRS_END_RANGE):

            setattr(proxy, f"dyn_attr{ind}", ind)

            if ind != 8:
                assert getattr(proxy, f"dyn_attr{ind}") == ind

        proxy.make_allowed(False)

        for ind in range(1, DYN_ATTRS_END_RANGE):

            if ind != 8:
                with pytest.raises(DevFailed):
                    _ = getattr(proxy, f"dyn_attr{ind}")

            with pytest.raises(DevFailed):
                setattr(proxy, f"dyn_attr{ind}", ind)


@pytest.mark.parametrize("use_green_mode", [True, False])
def test_dynamic_attribute_with_green_mode(use_green_mode, server_green_mode):
    class TestDevice(Device):
        green_mode = server_green_mode
        attr_value = 123

        def initialize_dynamic_attributes(self):
            global executor
            executor = get_executor(server_green_mode)
            attr = attribute(
                name="attr_r",
                dtype=int,
                access=AttrWriteType.READ,
                fget=self.user_read,
                read_green_mode=use_green_mode,
            )
            self.add_attribute(attr)
            attr = attribute(
                name="attr_rw",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget=self.user_read,
                fset=self.user_write,
                read_green_mode=use_green_mode,
                write_green_mode=use_green_mode,
            )
            self.add_attribute(attr)
            attr = attribute(
                name="attr_ia",
                dtype=int,
                access=AttrWriteType.READ,
                fget=self.user_read,
                fisallowed=self.user_is_allowed,
                read_green_mode=use_green_mode,
                isallowed_green_mode=use_green_mode,
            )
            self.add_attribute(attr)
            attr = attribute(
                name="attr_rw_always_ok",
                dtype=int,
                access=AttrWriteType.READ_WRITE,
                fget=self.user_read,
                fset=self.user_write,
                green_mode=True,
            )
            self.add_attribute(attr)

        sync_code = textwrap.dedent(
            """
            def user_read(self, attr):
                self.assert_executor_context_correct(attr.get_name())
                return self.attr_value

            def user_write(self, attr):
                self.assert_executor_context_correct(attr.get_name())
                self.attr_value = attr.get_write_value()

            def user_is_allowed(self, req_type):
                self.assert_executor_context_correct()
                assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
                return True

        """
        )

        def assert_executor_context_correct(self, attr_name=""):
            check_required = attr_name != "attr_rw_always_ok"
            if check_required and executor.asynchronous:
                assert executor.in_executor_context() == use_green_mode

        if server_green_mode == GreenMode.Asyncio and use_green_mode:
            exec(sync_code.replace("def", "async def"))
        else:
            exec(sync_code)

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.attr_r == 123
        proxy.attr_rw = 456
        assert proxy.attr_rw == 456
        assert proxy.attr_ia == 456


@pytest.mark.parametrize(
    "device_impl_class", [Device_4Impl, Device_5Impl, Device_6Impl, LatestDeviceImpl]
)
def test_dynamic_attribute_using_classic_api_like_sardana(device_impl_class):
    class ClassicAPIClass(DeviceClass):
        cmd_list = {
            "make_allowed": [[DevBoolean, "allow access"], [DevVoid, "none"]],
        }

        def __init__(self, name):
            super().__init__(name)
            self.set_type("TestDevice")

    class ClassicAPIDeviceImpl(device_impl_class):
        def __init__(self, cl, name):
            super().__init__(cl, name)
            ClassicAPIDeviceImpl.init_device(self)

        def init_device(self):
            self._attr1 = 3.14
            self._is_test_attr_allowed = True
            read = self.__class__._read_attr
            write = self.__class__._write_attr
            is_allowed = self.__class__._is_attr_allowed
            attr_name = "attr1"
            data_info = [[DevDouble, SCALAR, READ_WRITE]]
            dev_class = self.get_device_class()
            attr_data = AttrData(attr_name, dev_class.get_name(), data_info)
            self.add_attribute(attr_data, read, write, is_allowed)

        def _is_attr_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self._is_test_attr_allowed

        def _read_attr(self, attr):
            attr.set_value(self._attr1)

        def _write_attr(self, attr):
            w_value = attr.get_write_value()
            self._attr1 = w_value

        def make_allowed(self, yesno):
            self._is_test_attr_allowed = yesno

    with DeviceTestContext(ClassicAPIDeviceImpl, ClassicAPIClass) as proxy:
        proxy.make_allowed(True)
        assert proxy.attr1 == 3.14
        proxy.attr1 = 42.0
        assert proxy.attr1 == 42.0

        proxy.make_allowed(False)
        with pytest.raises(DevFailed):
            _ = proxy.attr1
        with pytest.raises(DevFailed):
            proxy.attr1 = 12.0


@pytest.mark.parametrize("read_function_signature", ["low_level", "high_level"])
@pytest.mark.parametrize("patched", [True, False])
def test_dynamic_attribute_with_unbound_functions(read_function_signature, patched):
    value = {"attr": None}
    is_allowed = None

    def low_level_read_function(device, attr):
        assert isinstance(device, TestDevice)
        attr.set_value(value["attr"])

    def high_level_read_function(device, attr):
        assert isinstance(device, TestDevice)
        return value["attr"]

    def write_function(device, attr):
        assert isinstance(device, TestDevice)
        value["attr"] = attr.get_write_value()

    def is_allowed_function(device, req_type):
        assert isinstance(device, TestDevice)
        assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
        return is_allowed

    class TestDevice(Device):

        def initialize_dynamic_attributes(self):
            if read_function_signature == "low_level":
                read_function = low_level_read_function
            elif read_function_signature == "high_level":
                read_function = high_level_read_function

            # trick to run server with non device method: patch __dict__
            if patched:
                self.__dict__["read_dyn_attr1"] = read_function
                self.__dict__["write_dyn_attr1"] = write_function
                self.__dict__["is_dyn_attr1_allowed"] = is_allowed_function
                attr = attribute(
                    name="dyn_attr1", dtype=int, access=AttrWriteType.READ_WRITE
                )
                self.add_attribute(attr)

                setattr(self, "read_dyn_attr2", read_function)
                setattr(self, "write_dyn_attr2", write_function)
                setattr(self, "is_dyn_attr2_allowed", is_allowed_function)
                attr = attribute(
                    name="dyn_attr2", dtype=int, access=AttrWriteType.READ_WRITE
                )
                self.add_attribute(attr)

            else:
                attr = attribute(
                    name="dyn_attr",
                    fget=read_function,
                    fset=write_function,
                    fisallowed=is_allowed_function,
                    dtype=int,
                    access=AttrWriteType.READ_WRITE,
                )
                self.add_attribute(attr)

    with DeviceTestContext(TestDevice) as proxy:
        is_allowed = True
        if patched:
            proxy.dyn_attr1 = 123
            assert proxy.dyn_attr1 == 123

            proxy.dyn_attr2 = 456
            assert proxy.dyn_attr2 == 456
        else:
            proxy.dyn_attr = 789
            assert proxy.dyn_attr == 789

        is_allowed = False
        if patched:
            with pytest.raises(DevFailed):
                proxy.dyn_attr1 = 123

            with pytest.raises(DevFailed):
                _ = proxy.dyn_attr1

            with pytest.raises(DevFailed):
                proxy.dyn_attr2 = 456

            with pytest.raises(DevFailed):
                _ = proxy.dyn_attr2
        else:
            with pytest.raises(DevFailed):
                proxy.dyn_attr = 123

            with pytest.raises(DevFailed):
                _ = proxy.dyn_attr


def test_attribute_decorators(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class BaseTestDevice(Device):
            @command(dtype_in=bool)
            async def make_allowed(self, yesno):
                self.is_allowed = yesno

    else:

        class BaseTestDevice(Device):
            @command(dtype_in=bool)
            def make_allowed(self, yesno):
                self.is_allowed = yesno

    class TestDevice(BaseTestDevice):
        green_mode = server_green_mode
        current_value = None
        voltage_value = None
        is_allowed = None

        current = attribute(label="Current", unit="mA", dtype=float)
        voltage = attribute(label="Voltage", unit="V", dtype=float)

        sync_code = textwrap.dedent(
            """
        @current.getter
        def cur_read(self):
            return self.current_value

        @current.setter
        def cur_write(self, current):
            self.current_value = current

        @current.is_allowed
        def cur_allo(self, req_type):
            return self.is_allowed

        @voltage.read
        def vol_read(self):
            return self.voltage_value

        @voltage.write
        def vol_write(self, voltage):
            self.voltage_value = voltage

        @voltage.is_allowed
        def vol_allo(self, req_type):
            return self.is_allowed
        """
        )

        if server_green_mode == GreenMode.Asyncio:
            exec(sync_code.replace("def ", "async def "))
        else:
            exec(sync_code)

    with DeviceTestContext(TestDevice) as proxy:
        proxy.make_allowed(True)
        proxy.current = 2.0
        assert_close(proxy.current, 2.0)
        proxy.voltage = 3.0
        assert_close(proxy.voltage, 3.0)

        proxy.make_allowed(False)
        with pytest.raises(DevFailed):
            proxy.current = 4.0
        with pytest.raises(DevFailed):
            _ = proxy.current
        with pytest.raises(DevFailed):
            proxy.voltage = 4.0
        with pytest.raises(DevFailed):
            _ = proxy.voltage


def test_attribute_info_description():
    class TestDevice(Device):
        @attribute(description="Description from kwarg has priority")
        def attr_description_kwarg(self) -> float:
            """Docstring has lower priority than description kwarg, so ignored"""
            return 1.5

        @attribute(access=AttrWriteType.READ_WRITE)
        def attr_decorated_docstring_only(self) -> float:
            """Docstring from decorated read method"""
            return 2.5

        @attr_decorated_docstring_only.setter
        def attr_decorated_docstring_only(self, value: float) -> None:
            """Docstring from decorated write method currently ignored"""
            pass

        attr_assignment = attribute()

        def read_attr_assignment(self) -> float:
            """Docstring from read method currently ignored for description"""
            return 3.5

    with DeviceTestContext(TestDevice) as proxy:
        cfg = proxy.get_attribute_config("attr_description_kwarg")
        assert cfg.description == "Description from kwarg has priority"

        cfg = proxy.get_attribute_config("attr_decorated_docstring_only")
        assert cfg.description == "Docstring from decorated read method"

        cfg = proxy.get_attribute_config("attr_assignment")
        assert cfg.description == tango.constants.DescNotSpec


def test_read_only_dynamic_attribute_with_dummy_write_method(
    dynamic_attribute_read_function,
):
    def dummy_write_method():
        return None

    class TestDevice(Device):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.attr_value = 123

        def initialize_dynamic_attributes(self):
            self.add_attribute(
                Attr("dyn_attr", DevLong, AttrWriteType.READ),
                r_meth=self.read_dyn_attr,
                w_meth=dummy_write_method,
            )

        def read_dyn_attr(self, attr):
            return self.attr_value

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.dyn_attr == 123


def test_dynamic_attribute_with_method_in_other_class():
    class Helper:
        value = 0
        is_allowed = True

        def _read_method(self, device, attr):
            assert isinstance(device, TestDevice)
            assert attr.get_name() == "dyn_attr"
            return self.value

        def _write_method(self, device, attr):
            assert isinstance(device, TestDevice)
            assert attr.get_name() == "dyn_attr"
            self.value = attr.get_write_value()

        def _is_allowed_method(self, device, req_type):
            assert isinstance(device, TestDevice)
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self.is_allowed

        def read_method(self, device, attr):
            return self._read_method(device, attr)

        def write_method(self, device, attr):
            self._write_method(device, attr)

        def is_allowed_method(self, device, req_type):
            return self._is_allowed_method(device, req_type)

    class TestDevice(Device):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.helper = Helper()

        def initialize_dynamic_attributes(self):
            self.add_attribute(
                Attr("dyn_attr", DevLong, AttrWriteType.READ_WRITE),
                r_meth=self.helper.read_method,
                w_meth=self.helper.write_method,
                is_allo_meth=self.helper.is_allowed_method,
            )

    with DeviceTestContext(TestDevice) as proxy:
        Helper.is_allowed = True
        proxy.dyn_attr = 123
        assert proxy.dyn_attr == 123

        Helper.is_allowed = False
        with pytest.raises(DevFailed):
            proxy.dyn_attr = 456
        with pytest.raises(DevFailed):
            _ = proxy.dyn_attr


def test_polled_attribute(server_green_mode):
    dct = {"PolledAttribute1": 100, "PolledAttribute2": 100000, "PolledAttribute3": 500}

    class TestDevice(Device):
        green_mode = server_green_mode

        @attribute(polling_period=dct["PolledAttribute1"])
        def PolledAttribute1(self):
            return 42.0

        @attribute(polling_period=dct["PolledAttribute2"])
        def PolledAttribute2(self):
            return 43.0

        @attribute(polling_period=dct["PolledAttribute3"])
        def PolledAttribute3(self):
            return 44.0

    with DeviceTestContext(TestDevice) as proxy:
        ans = proxy.polling_status()
        for x in ans:
            lines = x.split("\n")
            attr = lines[0].split("= ")[1]
            poll_period = int(lines[1].split("= ")[1])
            assert dct[attr] == poll_period


@pytest.mark.parametrize("return_type", [tuple, list])
@pytest.mark.parametrize("return_quality", [True, False])
def test_read_write_dev_encoded(dev_encoded_values, return_quality, return_type):

    def check_ans(raw_ans, expected_type):
        assert len(raw_ans) == 2
        assert is_pure_str(raw_ans[0])
        if expected_type is bytes:
            assert isinstance(raw_ans[1], bytes)
            assert raw_ans[1] == UTF8_STRING.encode()
        elif expected_type is str:
            assert is_pure_str(raw_ans[1])
            assert raw_ans[1] == UTF8_STRING
        else:
            assert isinstance(raw_ans[1], bytearray)
            assert raw_ans[1] == bytearray(UTF8_STRING.encode())

    class TestDevice(Device):
        attr_value = None
        command_value = None

        @attribute(dtype=DevEncoded, access=AttrWriteType.READ_WRITE)
        def attr(self):
            if return_quality:
                return return_type(self.attr_value), time.time(), AttrQuality.ATTR_VALID
            else:
                return return_type(self.attr_value)

        @attr.write
        def attr(self, value):
            check_ans(value, bytes)
            self.attr_value = value

        @attribute(dtype=DevEncoded)
        def attr_time_quality_1(self):
            return self.attr_value, 1.0, AttrQuality.ATTR_ALARM

        @attribute(dtype=DevEncoded)
        def attr_time_quality_2(self):
            return *self.attr_value, 2.0, AttrQuality.ATTR_WARNING

        @command(dtype_in=DevEncoded)
        def cmd_in(self, value):
            check_ans(value, bytes)
            self.command_value = value

        @command(dtype_out=DevEncoded)
        def cmd_out(self):
            return self.command_value

        @command(dtype_in=DevEncoded, dtype_out=DevEncoded)
        def cmd_in_out(self, value):
            check_ans(value, bytes)
            return value

    with DeviceTestContext(TestDevice) as proxy:
        proxy.attr = dev_encoded_values
        raw_ans = proxy.read_attribute("attr", extract_as=ExtractAs.Bytes)
        check_ans(raw_ans.value, bytes)
        check_ans(raw_ans.w_value, bytes)

        raw_ans = proxy.read_attribute("attr", extract_as=ExtractAs.String)
        check_ans(raw_ans.value, str)
        check_ans(raw_ans.w_value, str)

        raw_ans = proxy.read_attribute("attr", extract_as=ExtractAs.ByteArray)
        check_ans(raw_ans.value, bytearray)
        check_ans(raw_ans.w_value, bytearray)

        raw_ans = proxy.read_attribute(
            "attr_time_quality_1", extract_as=ExtractAs.Bytes
        )
        check_ans(raw_ans.value, bytes)
        assert raw_ans.quality == AttrQuality.ATTR_ALARM
        assert raw_ans.time.tv_sec == 1

        raw_ans = proxy.read_attribute(
            "attr_time_quality_2", extract_as=ExtractAs.Bytes
        )
        check_ans(raw_ans.value, bytes)
        assert raw_ans.quality == AttrQuality.ATTR_WARNING
        assert raw_ans.time.tv_sec == 2

        proxy.cmd_in(dev_encoded_values)
        raw_ans = proxy.cmd_out()
        check_ans(raw_ans, bytes)

        raw_ans = proxy.cmd_in_out(dev_encoded_values)
        check_ans(raw_ans, bytes)


def test_dev_encoded_wrong_encoding():
    class TestDevice(Device):
        @attribute(dtype=DevEncoded, access=AttrWriteType.READ)
        def attr_bad_encoding(self):
            return "utf-16", UTF8_STRING.encode("utf-16")

        @attribute(dtype=DevEncoded, access=AttrWriteType.READ)
        def attr_bad_string(self):
            return "bad_one", b"\xff"

    with DeviceTestContext(TestDevice) as proxy:
        with pytest.raises(UnicodeDecodeError):
            _ = proxy.read_attribute("attr_bad_encoding", extract_as=ExtractAs.String)

        with pytest.raises(UnicodeDecodeError):
            _ = proxy.read_attribute("attr_bad_string", extract_as=ExtractAs.String)


ENCODED_ATTRIBUTES = (  # encode function, decode function, data
    ("gray8", "gray8", np.array([np.arange(100, dtype=np.byte) for _ in range(5)])),
    ("gray16", "gray16", np.array([np.arange(100, dtype=np.int16) for _ in range(5)])),
    ("rgb24", "decode_24", np.ones((20, 10, 3), dtype=np.uint8)),
    (
        "jpeg_gray8",
        "gray8",
        np.array([np.arange(100, dtype=np.byte) for _ in range(5)]),
    ),
    ("jpeg_rgb24", None, np.ones((10, 10, 3), dtype=np.uint8)),
    ("jpeg_rgb32", "rgb32", np.zeros((10, 10, 3), dtype=np.uint8)),
)


def test_set_value_None():
    none_position = 0

    class TestDevice(Device):
        def init_device(self):
            super().init_device()
            attr = attribute(name="attr")
            self.add_attribute(attr, self._read_attr_that_raises)

            attr = attribute(name="attr_with_data_quality")
            self.add_attribute(attr, self._read_attr_with_date_quality_that_raises)

        def _read_attr_that_raises(self, attr):
            invalid_value = None
            attr.set_value(invalid_value)

        def _read_attr_with_date_quality_that_raises(self, attr):
            invalid_value = [1, 1, AttrQuality.ATTR_VALID]
            invalid_value[none_position] = None
            attr.set_value_date_quality(*invalid_value)

    with DeviceTestContext(TestDevice) as proxy:
        with pytest.raises(DevFailed, match="method cannot be called with None"):
            _ = proxy.attr

        for _ in range(3):
            with pytest.raises(DevFailed, match="method cannot be called with None"):
                _ = proxy.attr_with_data_quality
            none_position += 1


@pytest.mark.parametrize(
    "f_encode, f_decode, data",
    ENCODED_ATTRIBUTES,
    ids=[f_name for f_name, _, _ in ENCODED_ATTRIBUTES],
)
def test_encoded_attribute(f_encode, f_decode, data):
    if f_encode == "jpeg_rgb32":
        pytest.xfail(
            "jpeg_rgb32 needs cppTango built with TANGO_USE_JPEG option, so we skip this test"
        )

    class TestDevice(Device):
        @attribute(dtype=DevEncoded, access=AttrWriteType.READ)
        def attr(self):
            enc = EncodedAttribute()
            getattr(enc, f"encode_{f_encode}")(data)
            return enc

    def decode_24(data_to_decode):
        # first two bytes are width, then two bytes are height and then goes encoded image
        # see implementation of other decode methods, e.g.
        # https://gitlab.com/tango-controls/cppTango/-/blob/main/cppapi/server/encoded_attribute.cpp#L229
        width = int.from_bytes(data_to_decode[:2], byteorder="big")
        height = int.from_bytes(data_to_decode[2:4], byteorder="big")
        decoded_data = np.frombuffer(data_to_decode[4:], dtype=np.uint8)
        return decoded_data.reshape((height, width, 3))

    with DeviceTestContext(TestDevice) as proxy:
        if f_decode == "decode_24":
            ret = proxy.read_attribute("attr", extract_as=ExtractAs.Bytes)
            codec, ret_data = ret.value
            assert_close(decode_24(ret_data), data)
        elif f_decode is not None:
            ret = proxy.read_attribute("attr", extract_as=ExtractAs.Nothing)
            enc = EncodedAttribute()
            assert_close(getattr(enc, f"decode_{f_decode}")(ret), data)
        else:
            # for jpeg24 we test only encode
            pass


@pytest.mark.xfail(reason="Somehow this test became too fragile, need better solution")
def test_dev_encoded_memory_usage():
    LARGE_DATA_SIZE = 10 * 1024 * 1024  # 1 Mb should be enough, but 10 more reliable
    NUM_CYCLES = 5

    def check_ans(raw_ans):
        assert len(raw_ans) == 2
        assert is_pure_str(raw_ans[0])
        assert isinstance(raw_ans[1], bytes)
        assert len(raw_ans[1]) == LARGE_DATA_SIZE
        if raw_ans[0] == "str":
            assert raw_ans[1] == b"a" * LARGE_DATA_SIZE
        elif raw_ans[0] == "bytes":
            assert raw_ans[1] == b"b" * LARGE_DATA_SIZE
        else:
            assert raw_ans[1] == b"c" * LARGE_DATA_SIZE

    class TestDevice(Device):
        attr_str_read = attribute(
            dtype=DevEncoded, access=AttrWriteType.READ, fget="read_str"
        )
        attr_str_write = attribute(
            dtype=DevEncoded, access=AttrWriteType.WRITE, fset="write_str"
        )
        attr_str_read_write = attribute(
            dtype=DevEncoded,
            access=AttrWriteType.READ_WRITE,
            fget="read_str",
            fset="write_str",
        )

        attr_bytes_read = attribute(
            dtype=DevEncoded, access=AttrWriteType.READ, fget="read_bytes"
        )
        attr_bytes_write = attribute(
            dtype=DevEncoded, access=AttrWriteType.WRITE, fset="write_bytes"
        )
        attr_bytes_read_write = attribute(
            dtype=DevEncoded,
            access=AttrWriteType.READ_WRITE,
            fget="read_bytes",
            fset="write_bytes",
        )

        attr_bytearray_read = attribute(
            dtype=DevEncoded, access=AttrWriteType.READ, fget="read_bytearray"
        )
        attr_bytearray_write = attribute(
            dtype=DevEncoded, access=AttrWriteType.WRITE, fset="write_bytearray"
        )
        attr_bytearray_read_write = attribute(
            dtype=DevEncoded,
            access=AttrWriteType.READ_WRITE,
            fget="read_bytearray",
            fset="write_bytearray",
        )

        def read_str(self):
            return "str", "a" * LARGE_DATA_SIZE

        def write_str(self, value):
            check_ans(value)

        def read_bytes(self):
            return "bytes", b"b" * LARGE_DATA_SIZE

        def write_bytes(self, value):
            check_ans(value)

        def read_bytearray(self):
            return "bytearray", bytearray(b"c" * LARGE_DATA_SIZE)

        def write_bytearray(self, value):
            check_ans(value)

        @command(dtype_in=DevEncoded, dtype_out=DevEncoded)
        def cmd_in_out(self, value):
            check_ans(value)
            if value[0] == "str":
                return "str", "a" * LARGE_DATA_SIZE
            elif value[0] == "bytes":
                return "bytes", b"b" * LARGE_DATA_SIZE
            else:
                return "bytearray", bytearray(b"c" * LARGE_DATA_SIZE)

    with DeviceTestContext(TestDevice) as proxy:
        last_memory_usage = []
        for cycle in range(NUM_CYCLES):
            proxy.attr_str_write = "str", "a" * LARGE_DATA_SIZE
            proxy.attr_bytes_write = "bytes", b"b" * LARGE_DATA_SIZE
            proxy.attr_bytearray_write = "bytearray", bytearray(b"c" * LARGE_DATA_SIZE)

            proxy.attr_str_read_write = "str", "a" * LARGE_DATA_SIZE
            proxy.attr_bytes_read_write = "bytes", b"b" * LARGE_DATA_SIZE
            proxy.attr_bytearray_read_write = "bytearray", bytearray(
                b"c" * LARGE_DATA_SIZE
            )

            check_ans(proxy.attr_str_read)
            check_ans(proxy.attr_bytes_read)
            check_ans(proxy.attr_bytearray_read)

            check_ans(proxy.attr_str_read_write)
            check_ans(proxy.attr_bytes_read_write)
            check_ans(proxy.attr_bytearray_read_write)

            check_ans(proxy.cmd_in_out(("str", "a" * LARGE_DATA_SIZE)))
            check_ans(proxy.cmd_in_out(("bytes", b"b" * LARGE_DATA_SIZE)))
            check_ans(
                proxy.cmd_in_out(("bytearray", bytearray(b"c" * LARGE_DATA_SIZE)))
            )

            current_memory_usage = int(psutil.Process(os.getpid()).memory_info().rss)
            last_memory_usage = np.append(last_memory_usage, current_memory_usage)
        assert not np.all(np.diff(last_memory_usage) > 0)


def test_attribute_list():
    class TestDevice(Device):

        _val = 0

        @attribute
        def attr(self) -> int:
            return self._val

        @attr.write
        def set_attr(self, val_in: int):
            self._val = val_in

        @attribute
        def attr2(self) -> str:
            return "a"

        @command()
        def check_attribute_list(self):
            attribute_list = self.get_device_attr().get_attribute_list()
            assert isinstance(attribute_list, tuple)
            assert len(attribute_list) == 4

            assert isinstance(attribute_list[0], WAttribute)
            assert attribute_list[0].get_name() == "attr"

            assert isinstance(attribute_list[1], Attribute)
            assert attribute_list[1].get_name() == "attr2"

        @command()
        def check_attr_list(self):
            u = Util.instance()

            classes = u.get_class_list()
            assert len(classes) == 1
            assert isinstance(classes[0], DeviceClass)
            multi_class_attribute = classes[0].get_class_attr()
            assert isinstance(multi_class_attribute, MultiClassAttribute)

            attr_list = multi_class_attribute.get_attr_list()
            assert isinstance(attr_list, tuple)

            assert len(attr_list) == 2

            assert isinstance(attr_list[0], Attr)
            assert attr_list[0].get_name() == "attr"

            assert isinstance(attr_list[1], Attr)
            assert attr_list[1].get_name() == "attr2"

    with DeviceTestContext(TestDevice) as proxy:
        proxy.check_attribute_list()
        proxy.check_attr_list()


@pytest.mark.parametrize("set_w_value", [False, True])
def test_fill_attr_polling_buffer(attribute_typed_values, set_w_value):
    dtype, values, expected = attribute_typed_values

    start_time = time.time()

    class TestDevice(Device):

        @attribute(
            dtype=dtype, access=AttrWriteType.READ_WRITE, max_dim_x=3, max_dim_y=3
        )
        def attr(self):
            return values[0]

        @attr.write
        def attr(self, new_val):
            pass

        @attribute(dtype=dtype, max_dim_x=3, max_dim_y=3)
        def attr_2(self):
            return values[0]

        @command
        def fill_history(self):
            data = []
            for i, val in enumerate(values):
                t = start_time + i
                w_val = val if set_w_value else None
                data.append(TimedAttrData(val, w_value=w_val, time_stamp=t))

            data.append(
                TimedAttrData(
                    error=RuntimeError("Test"),
                    time_stamp=start_time + len(data),
                )
            )
            self.fill_attr_polling_buffer("attr", data)

            # check how auto-convert to list works
            self.fill_attr_polling_buffer("attr_2", TimedAttrData(values[0]))

    with DeviceTestContext(TestDevice) as proxy:
        # we do not want, that tango core fills up history automatically
        proxy.poll_attribute("attr", 0)
        proxy.poll_attribute("attr_2", 0)
        proxy.fill_history()
        history = proxy.attribute_history("attr", 10)
        assert len(history) == len(values) + 1
        for ind, attr in enumerate(history[:-1]):
            assert_close(attr.value, expected(values[ind]))
            assert attr.quality == AttrQuality.ATTR_VALID
            assert not attr.has_failed
            assert_close(start_time + ind, attr.time.totime())
            if set_w_value:
                assert_close(attr.w_value, expected(values[ind]))

        assert history[-1].has_failed
        assert history[-1].quality == AttrQuality.ATTR_INVALID

        history = proxy.attribute_history("attr_2", 10)
        assert_close(history[0].value, expected(values[0]))


def test_removed_dim_parameters():
    reason = "Note, that dim_x and dim_y arguments are no longer supported"

    class TestDevice(Device):
        def initialize_dynamic_attributes(self):
            attr = attribute(name="attr", fget=self.attr, dtype=int)
            self.add_attribute(attr)

        def attr(self, attr):
            with pytest.raises(TypeError, match=reason):
                attr.set_value_date_quality(1, time.time(), AttrQuality.ATTR_VALID, 1)
            with pytest.raises(TypeError, match=reason):
                attr.set_value_date_quality(
                    1, time.time(), AttrQuality.ATTR_VALID, 1, 1
                )
            with pytest.raises(TypeError, match=reason):
                attr.set_value(1, 1)
            with pytest.raises(TypeError, match=reason):
                attr.set_value(1, 1, 2)

            return 1

    with DeviceTestContext(TestDevice) as proxy:
        _ = proxy.attr



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

# Imports
import sys
import os
import functools
from collections import namedtuple

# Gevent imports
import gevent.event
import gevent.queue
import gevent.monkey
import gevent.threadpool

# Bypass gevent monkey patching
ThreadSafeEvent = gevent.monkey.get_original("threading", "Event")

# Tango imports
from tango.green import AbstractExecutor, get_ident
from tango.utils import _get_current_otel_context, _get_non_tango_source_location

__all__ = (
    "GeventExecutor",
    "get_global_executor",
    "set_global_executor",
    "_switch_global_executor_to_thread",
)

# Global executor

_MAIN_EXECUTOR = None
_THREAD_POOL = None
_THREAD_EXECUTORS = {}


def _switch_global_executor_to_thread():
    """
    internal PyTango function, use only if you sure, what you are doing!
    Used for correct behavior of TestDeviceContext
    checks, that global executor belongs to the caller thread, and,
    if not - creates a new one and saves it as a new global
    """
    global _MAIN_EXECUTOR
    if _MAIN_EXECUTOR is not None and not _MAIN_EXECUTOR.in_executor_context():
        # we save current executor in the known subthread executors to be used later
        _THREAD_EXECUTORS[_MAIN_EXECUTOR.get_ident()] = _MAIN_EXECUTOR
        _MAIN_EXECUTOR = GeventExecutor(subexecutor=ThreadPool(maxsize=100))


def get_global_executor():
    global _MAIN_EXECUTOR
    if _MAIN_EXECUTOR is None:
        _MAIN_EXECUTOR = GeventExecutor()

    # the following patch is used for correct behavior of TestDeviceContext,
    # which has two different executors for main and device threads
    if not _MAIN_EXECUTOR.in_executor_context():
        ident = get_ident(), os.getpid()
        if ident in _THREAD_EXECUTORS:
            return _THREAD_EXECUTORS[ident]

    return _MAIN_EXECUTOR


def set_global_executor(executor):
    global _MAIN_EXECUTOR
    _MAIN_EXECUTOR = executor


def get_global_threadpool():
    global _THREAD_POOL
    if _THREAD_POOL is None:
        _THREAD_POOL = ThreadPool(maxsize=10**4)
    return _THREAD_POOL


ExceptionInfo = namedtuple("ExceptionInfo", "type value traceback")


def wrap_error(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            return ExceptionInfo(*sys.exc_info())

    return wrapper


def unwrap_error(source):
    destination = gevent.event.AsyncResult()

    def link(source):
        if isinstance(source.value, ExceptionInfo):
            try:
                destination.set_exception(source.value.value, exc_info=source.value)
            # Gevent 1.0 compatibility
            except TypeError:
                destination.set_exception(source.value.value)
            return
        destination(source)

    source.rawlink(link)
    return destination


class ThreadPool(gevent.threadpool.ThreadPool):
    def spawn(self, fn, *args, **kwargs):
        wrapped = wrap_error(fn)
        raw = super().spawn(wrapped, *args, **kwargs)
        return unwrap_error(raw)


# Gevent task and event loop


class GeventTask:
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.value = None
        self.exception = None
        self.done = ThreadSafeEvent()
        self.started = ThreadSafeEvent()

    def run(self):
        self.started.set()
        try:
            self.value = self.func(*self.args, **self.kwargs)
        except Exception:
            self.exception = sys.exc_info()
        finally:
            self.done.set()

    def spawn(self):
        return gevent.spawn(self.run)

    def result(self):
        self.done.wait()
        if self.exception:
            raise self.exception[1]
        return self.value


# Gevent executor


class GeventExecutor(AbstractExecutor):
    """Gevent tango executor"""

    asynchronous = True
    default_wait = True

    def __init__(self, loop=None, subexecutor=None):
        super().__init__()
        if loop is None:
            loop = gevent.get_hub().loop
        if subexecutor is None:
            subexecutor = get_global_threadpool()
        self.loop = loop
        self.subexecutor = subexecutor

    def delegate(self, fn, *args, **kwargs):
        """Return the given operation as a gevent future."""
        if hasattr(fn, "__trace_kwargs__"):
            kwargs["trace_location"] = _get_non_tango_source_location()
            kwargs["trace_context"] = _get_current_otel_context()
        return self.subexecutor.spawn(fn, *args, **kwargs)

    def access(self, accessor, timeout=None):
        """Return a result from an gevent future."""
        return accessor.get(timeout=timeout)

    def create_watcher(self):
        try:
            return self.loop.async_()
        except AttributeError:
            return getattr(self.loop, "async")()

    def submit(self, fn, *args, **kwargs):
        task = GeventTask(fn, *args, **kwargs)
        watcher = self.create_watcher()
        watcher.start(task.spawn)
        watcher.send()
        task.started.wait()
        # The watcher has to be stopped in order to be garbage-collected.
        # This step is crucial since the watcher holds a reference to the
        # `task.spawn` method which itself holds a reference to the task.
        # It's also important to wait for the task to be spawned before
        # stopping the watcher, otherwise the task won't run.
        watcher.stop()
        return task

    def execute(self, fn, *args, **kwargs):
        """Execute an operation and return the result."""
        if self.in_executor_context():
            return fn(*args, **kwargs)
        task = self.submit(fn, *args, **kwargs)
        return task.result()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
Power supply device that connects to a "hardware" simulator
provided by ps-simulator.py.  The communication is handled
by a library, instead of the Tango device directly.
Launch the ps-simulator.py script before starting this device server.
"""

from tango.server import Device, attribute, command, device_property

from libps import PowerSupplyClient


class PowerSupply(Device):
    host = device_property(str, default_value="localhost")
    port = device_property(int, default_value=45000)

    def init_device(self):
        super().init_device()
        self.ps_client = PowerSupplyClient(self.host, self.port)

    def delete_device(self):
        self.ps_client.disconnect()

    @attribute(dtype=float)
    def voltage(self):
        return self.ps_client.get_voltage()

    @command
    def calibrate(self):
        self.ps_client.do_calibration()


if __name__ == "__main__":
    PowerSupply.run_server()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import pathlib
import sys
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from argparse import ArgumentParser


@dataclass
class Frame:
    index: int
    function: str
    directory: str
    filename: str
    line_no: int
    ip: str
    object: str

    @classmethod
    def from_element(cls, index, frame: ET.Element):
        return cls(
            index=index,
            function=get_element_text(frame, "fn"),
            directory=get_element_text(frame, "dir"),
            filename=get_element_text(frame, "file"),
            line_no=int(get_element_text(frame, "line", "0")),
            ip=get_element_text(frame, "ip"),
            object=get_element_text(frame, "obj", "???"),
        )

    def __str__(self):
        prefix = "at" if self.index == 0 else "by"
        function = short_function(self.function)
        object = short_path(self.object)
        source = f"{short_path(self.directory, self.filename)}:{self.line_no}"
        if function:
            return f"{prefix} {self.ip:>10}: {function} ({source} in {object})"
        else:
            return f"{prefix} {self.ip:>10}: ???"


def short_function(function):
    if len(function) > 100:
        function = f"{function[:40]}...{function[-40:]}"
    return function


def short_path(*paths):
    path = pathlib.Path(*paths)
    interesting_parts = path.parts[-4:]
    return pathlib.Path("...").joinpath(*interesting_parts)


@dataclass
class Stack:
    frames: list[Frame]

    @classmethod
    def from_element(cls, stack: ET.Element):
        frames = []
        for index, frame in enumerate(stack.findall("frame")):
            frame_data = Frame.from_element(index, frame)
            frames.append(frame_data)
        return cls(frames)


@dataclass
class XWhat:
    text: str
    leaked_bytes: int
    leaked_blocks: int

    @classmethod
    def from_element(cls, xwhat: ET.Element):
        return cls(
            text=get_element_text(xwhat, "text"),
            leaked_bytes=int(get_element_text(xwhat, "leakedbytes", "0")),
            leaked_blocks=int(get_element_text(xwhat, "leakedblocks", "0")),
        )


@dataclass
class LeakError:
    kind: str
    stack: Stack
    xwhat: XWhat

    @classmethod
    def from_element(cls, error: ET.Element):
        kind = error.find("kind")
        return cls(
            kind=kind.text,
            stack=Stack.from_element(error.find("stack")),
            xwhat=XWhat.from_element(error.find("xwhat")),
        )

    def __lt__(self, other):
        if not isinstance(other, LeakError):
            return NotImplemented
        return self.xwhat.leaked_bytes < other.xwhat.leaked_bytes


def get_element_text(parent: ET.Element, name: str, default: str = "") -> str:
    if parent is None:
        return default

    element = parent.find(name)
    if element is not None:
        return element.text
    else:
        return default


def get_leaks(root: ET.Element, kind: str) -> list[LeakError]:
    leaks = []
    for error_element in root.findall("error"):
        kind_element = error_element.find("kind")
        if kind_element is not None and kind_element.text == kind:
            leak = LeakError.from_element(error_element)
            leaks.append(leak)
    return leaks


def get_pytango_leaks(leaks: list[LeakError], max_blocks: int) -> list[LeakError]:
    pytango_leaks = []
    for leak in leaks:
        if leak.xwhat.leaked_blocks > max_blocks:
            if is_pytango_in_stack(leak):
                pytango_leaks.append(leak)
    return pytango_leaks


def is_pytango_in_stack(leak) -> bool:
    for frame in leak.stack.frames:
        filename = pathlib.Path(frame.object).name
        if filename.startswith("_tango") and filename.endswith(".so"):
            return True
    return False


def parse_input_file(filename: str) -> tuple[str, list[LeakError], list[LeakError]]:
    root = ET.parse(filename).getroot()
    header = get_header(root)
    definite_leaks = get_leaks(root, "Leak_DefinitelyLost")
    possible_leaks = get_leaks(root, "Leak_PossiblyLost")
    return header, definite_leaks, possible_leaks


def get_header(root: ET.Element) -> str:
    pid = root.find("pid")
    if pid is None:
        return ""
    ppid = root.find("ppid")
    if ppid is None:
        return ""
    args = root.find("args")
    if args is not None:
        argv = args.find("argv")
        if argv is not None:
            exe = argv.find("exe")
            if exe is None:
                return ""
            arg_list = []
            for arg in argv.findall("arg"):
                arg_list.append(arg.text)

    return f"ppid {ppid.text} -> pid {pid.text}: {exe.text} {' '.join(arg_list)}"


def report_pytango_leaks(label, leaks, max_blocks):
    pytango_leaks = get_pytango_leaks(leaks, max_blocks)
    print(
        f"PyTango {label} leaks larger than {max_blocks} block(s): {len(pytango_leaks)}"
    )
    print_leaks(pytango_leaks)
    return len(pytango_leaks)


def print_leaks(leaks: list[LeakError]):
    for count, leak in enumerate(sorted(leaks, reverse=True)):
        print(f"\n  Leak {count+1}:  {leak.xwhat.text}")
        for frame in leak.stack.frames:
            print(f"    {frame}")


def parse_args():
    parser = ArgumentParser()
    parser.add_argument("file", help="Valgrind XML file to parse")
    parser.add_argument(
        "--max-blocks",
        help="Report leaks larger than this many blocks",
        type=int,
        default=1,
    )
    parser.add_argument(
        "--no-error", action="store_true", help="Exit code 0, even if leaks found"
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    header, definite_leaks, possible_leaks = parse_input_file(args.file)
    print(
        f"Valgrind {header}:\n"
        f"\tfound {len(possible_leaks)} possible leaks, "
        f"and {len(definite_leaks)} definite leaks.\n"
    )

    max_blocks = args.max_blocks
    report_pytango_leaks("possible", possible_leaks, max_blocks)
    n_definite = report_pytango_leaks("definite", definite_leaks, max_blocks)

    # only definite leaks can cause an error on exit
    if n_definite > 0 and not args.no_error:
        print(f"\nError: definite leaks found in {args.file}!")
        sys.exit(1)
    else:
        print("\nDone")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

import threading
import time

from tango import (
    AttrWriteType,
    AttributeProxy,
    Database,
    DeviceProxy,
    EnsureOmniThread,
    EventType,
    InfoIt,
    Group,
)
from tango.test_context import MultiDeviceTestContext
from tango.server import Device, attribute, command, device_property
from tango.utils import EventCallback

# <<<<<< If user wants to trace their own things -------
# Dependencies:
# pip or conda install the following libraries:
#   opentelemetry-api opentelemetry-sdk opentelemetry-exporter-otlp-proto-grpc

# API
from opentelemetry import trace as trace_api

# SDK
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import (
    SERVICE_NAME,
    SERVICE_NAMESPACE,
    Resource,
)
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

resource = Resource.create({SERVICE_NAMESPACE: "org.institute", SERVICE_NAME: "my.app"})

provider = TracerProvider(resource=resource)
processor = BatchSpanProcessor(OTLPSpanExporter())
provider.add_span_processor(processor)

# Sets the global default tracer provider
trace_api.set_tracer_provider(provider)

# Creates a tracer from the global tracer provider
tracer = trace_api.get_tracer("user.tracer")
# ---- end of user's own tracing. >>>>>>

# import to pass context to a user thread:
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator


class Leader(Device):
    FollowerTRLs = device_property(dtype=(str,))

    @command(dtype_in=int)
    @InfoIt(show_args=True)
    def TurnFollowerOn(self, follower_id):
        self.debug_stream(f"Turning follower {follower_id} on...")
        follower_trl = self.FollowerTRLs[follower_id - 1]
        follower_device = DeviceProxy(follower_trl)
        follower_device.isOn = True

    @command(dtype_in=int)
    def TurnFollowerOff(self, follower_id):
        device_tracer = self.get_telemetry_tracer()
        with device_tracer.start_as_current_span("test.leader.span") as span:
            span.set_attribute("follower_id", follower_id)
            follower_trl = self.FollowerTRLs[follower_id - 1]
            follower_device = DeviceProxy(follower_trl)
            follower_device.isOn = False

    @command(dtype_in=bool, dtype_out=bool)
    def SetDeviceTracing(self, enable):
        self.set_telemetry_enabled(enable)
        return self.is_telemetry_enabled()

    @command(dtype_in=bool, dtype_out=bool)
    def SetKernelTracing(self, enable):
        self.set_kernel_tracing_enabled(enable)
        return self.is_kernel_tracing_enabled()

    @command(dtype_in=int)
    @InfoIt(show_args=True)
    def PollFollowerOnOff(self, follower_id):
        follower_trl = self.FollowerTRLs[follower_id - 1]
        follower_device = DeviceProxy(follower_trl)
        user_test_polling_and_events(follower_device)


class Follower(Device):
    def init_device(self):
        super().init_device()
        self._is_on = False

        attr = attribute(
            name="dynamicAttribute",
            access=AttrWriteType.READ,
            fget=self.read_dyn_attr,
        )
        self.add_attribute(attr)
        cmd = command(
            dtype_in=int,
            f=self.DynamicCommand,
        )
        self.add_command(cmd)

    isOn = attribute(access=AttrWriteType.READ_WRITE)

    @InfoIt(show_ret=True)
    def read_isOn(self) -> bool:
        return self._is_on

    @InfoIt(show_args=True)
    def write_isOn(self, value: bool) -> None:
        self._is_on = value

    def read_dyn_attr(self, arg) -> int:
        print(f"Follower.dynamicAttribute")
        return 123

    def DynamicCommand(self, arg):
        print(f"Follower.DynamicCommand({arg})")

    def delete_device(self):
        # just to show trace emitted for user code
        pass

    def dev_state(self):
        # just to show trace emitted for user code
        return super().dev_state()


devices_info = [
    {
        "class": Leader,
        "devices": (
            {
                "name": "device/leader/1",
                "properties": {
                    "FollowerTRLs": ["device/follower/1", "device/follower/2"],
                },
            },
        ),
    },
    {
        "class": Follower,
        "devices": [
            {"name": "device/follower/1", "properties": {}},
            {"name": "device/follower/2", "properties": {}},
        ],
    },
]


def call_from_thread_with_context(follower: DeviceProxy):
    carrier = {}
    TraceContextTextMapPropagator().inject(carrier)
    print(f"call_from_thread_with_context {carrier=}")
    thread = threading.Thread(target=thread_worker, args=(carrier, follower))
    thread.start()
    time.sleep(0.001)


def thread_worker(carrier, follower):
    with EnsureOmniThread():
        print(f"worker started for: {follower}")
        ctx = TraceContextTextMapPropagator().extract(carrier=carrier)
        with tracer.start_as_current_span("thread.worker.call", context=ctx):
            user_turn_follower_off_directly(follower)


def user_turn_follower_off_directly(follower: DeviceProxy):
    print(f"user_turn_follower_off_directly: {follower}")
    follower.isOn = False


def user_test_polling_and_events(follower: DeviceProxy):
    print(f"user_toggle_polling {follower.is_attribute_polled('isOn')=}")
    follower.poll_attribute("isOn", 10)
    print(f"user_toggle_polling {follower.is_attribute_polled('isOn')=}")
    user_test_events(follower)
    follower.stop_poll_attribute("isOn")
    print(f"user_toggle_polling {follower.is_attribute_polled('isOn')=}")


def user_test_events(follower: DeviceProxy):
    eid = follower.subscribe_event("isOn", EventType.CHANGE_EVENT, EventCallback())
    follower.isOn = False
    time.sleep(0.01)
    follower.isOn = True
    time.sleep(0.01)
    follower.unsubscribe_event(eid)


def main():
    with MultiDeviceTestContext(devices_info, process=False, timeout=600, debug=3):
        leader = DeviceProxy("device/leader/1")
        follower_1 = DeviceProxy("device/follower/1")
        follower_2 = DeviceProxy("device/follower/2")
        # user could create their own span, e.g.:
        with tracer.start_as_current_span(
            "my.app.main", kind=trace_api.SpanKind.CLIENT
        ):
            for loop in range(2):
                with tracer.start_as_current_span(
                    "my.app.main.inner-loop", kind=trace_api.SpanKind.CLIENT
                ) as span:
                    span.set_attribute("operation.value", loop)

                    # tell leader to enable both followers
                    leader.command_inout("TurnFollowerOn", 1)
                    leader.command_inout("TurnFollowerOn", 2)

                    # check initial state: both followers are on
                    _ = follower_1.read_attribute("isOn").value
                    _ = follower_2.isOn

                    # turn off, using low-level and high-level API
                    follower_1.write_attribute("isOn", 0)
                    user_turn_follower_off_directly(follower_2)

                    call_from_thread_with_context(follower_2)

                    leader.TurnFollowerOff(1)  # FIXME

                    with tracer.start_as_current_span(
                        "test.deviceproxy.other", kind=trace_api.SpanKind.CLIENT
                    ):
                        # read multiple attributes, and check config
                        _ = follower_2.read_attributes(["isOn", "state"])
                        _ = follower_1.get_attribute_config("isOn")
                        _ = follower_1.read_attribute("dynamicAttribute")
                        _ = follower_1.command_inout("DynamicCommand", 22)

                    with tracer.start_as_current_span(
                        "test.polling-and-events", kind=trace_api.SpanKind.CLIENT
                    ):
                        leader.PollFollowerOnOff(1)
                        # polling changes not within device context aren't traced yet
                        follower_1.poll_attribute("isOn", 1000)
                        follower_1.stop_poll_attribute("isOn")

                    with tracer.start_as_current_span(
                        "test.attributeproxy", kind=trace_api.SpanKind.CLIENT
                    ):
                        follower_1_is_on = AttributeProxy("device/follower/1/isOn")
                        _ = follower_1_is_on.ping()
                        reading = follower_1_is_on.read()
                        print(f"AttributeProxy reading: {reading.value}")

                    with tracer.start_as_current_span(
                        "test.group", kind=trace_api.SpanKind.CLIENT
                    ):
                        group = Group("add-one-at-a-time")
                        group.add("device/follower/1")
                        group.add("device/follower/2")
                        group.ping()
                        reply = group.read_attribute("isOn")
                        print(
                            f"Group reading: 1:{reply[0].get_data().value}, 2:{reply[1].get_data().value}"
                        )

                    with tracer.start_as_current_span(
                        "test.database", kind=trace_api.SpanKind.CLIENT
                    ):
                        db = Database()
                        _ = db.get_server_list()
                        _ = db.put_device_property("sys/tg_test/1", {"foo": "bar"})
                        prop_val = db.get_device_property("sys/tg_test/1", "foo")
                        print(f"got property foo: {prop_val}")
                        _ = db.delete_device_property("sys/tg_test/1", "foo")


if __name__ == "__main__":
    main()

    # instead of running main, can run as Tango server providing the two classes:
    # tango.server.run([Leader, Follower])

"""
Examples:

*** Change code above to use main() instead of tango.server.run(...)

# Run example with telemetry on, traces go to stdout
$ TANGO_TELEMETRY_ENABLE=on python prototyping.py

# Run example with telemetry on, traces to to local collector via gRPC
$ TANGO_TELEMETRY_ENABLE=on TANGO_TELEMETRY_TRACES_EXPORTER=grpc python

*** Change code above to use tango.server.run(...) instead of main()

# Python client with telemetry on, traces go to stdout
#  (create your own DeviceProxy)
$ TANGO_TELEMETRY_ENABLE=on python

# Python client with telemetry on, traces to to local collector via gRPC
# (create your own DeviceProxy)
$ TANGO_TELEMETRY_ENABLE=on TANGO_TELEMETRY_TRACES_EXPORTER=grpc python

# Run Leader device server, telemetry disabled
$ python prototyping.py Leader --host=127.0.0.1 -v3

# Run Leader device server, telemetry on, traces and logs to local collector via gRPC
$ TANGO_TELEMETRY_ENABLE=on TANGO_TELEMETRY_TRACES_EXPORTER=grpc TANGO_TELEMETRY_LOGS_EXPORTER=grpc python prototyping.py Leader --host=127.0.0.1 -v3

# Run Follower device server, telemetry on, traces and logs to local collector via gRPC
# Also include additional process info in the traces using experimental OpenTelemetry
# resource detector.
$ TANGO_TELEMETRY_ENABLE=on TANGO_TELEMETRY_TRACES_EXPORTER=grpc TANGO_TELEMETRY_LOGS_EXPORTER=grpc OTEL_EXPERIMENTAL_RESOURCE_DETECTORS=process python prototyping.py Follower --host=127.0.0.1 -v3

"""



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import os
import time
import warnings

import numpy as np
import pytest
import tango
import asyncio
import gevent.event
import concurrent.futures

from tango import AttrWriteType, DevFailed, GreenMode, get_device_proxy, DeviceProxy
from tango.asyncio_executor import AsyncioExecutor
from tango.device_server import get_worker
from tango.gevent_executor import GeventExecutor
from tango.green import SynchronousExecutor
from tango.server import Device
from tango.server import command, attribute, device_property
from tango.test_utils import (
    DeviceTestContext,
    MultiDeviceTestContext,
    SimpleDevice,
    ClassicAPISimpleDeviceImpl,
    ClassicAPISimpleDeviceClass,
)


WINDOWS = "nt" in os.name


class Device1(Device):
    _attr1 = 100

    def init_device(self):
        super().init_device()
        self.set_state(tango.DevState.ON)

    @attribute(dtype=int, access=AttrWriteType.READ_WRITE)
    def attr1(self):
        return self._attr1

    def write_attr1(self, value):
        self._attr1 = value


class Device1GreenModeUnspecified(Device1):
    pass


class Device1Synchronous(Device1):
    green_mode = GreenMode.Synchronous


class Device1Gevent(Device1):
    green_mode = GreenMode.Gevent


class Device1Asyncio(Device):
    green_mode = GreenMode.Asyncio

    _attr1 = 100

    async def init_device(self):
        await super().init_device()
        self.set_state(tango.DevState.ON)

    @attribute(dtype=int, access=AttrWriteType.READ_WRITE)
    async def attr1(self):
        return self._attr1

    async def write_attr1(self, value):
        self._attr1 = value


class Device2(Device):
    _attr2 = 200

    def init_device(self):
        super().init_device()
        self.set_state(tango.DevState.ON)

    @attribute
    def attr2(self):
        return self._attr2


class Device2GreenModeUnspecified(Device2):
    pass


class Device2Synchronous(Device2):
    green_mode = GreenMode.Synchronous


class Device2Gevent(Device2):
    green_mode = GreenMode.Gevent


class Device2Asyncio(Device):
    green_mode = GreenMode.Asyncio

    _attr2 = 200

    async def init_device(self):
        await super().init_device()
        self.set_state(tango.DevState.ON)

    @attribute
    async def attr2(self):
        return self._attr2


class Device3(Device):
    def init_device(self):
        super().init_device()
        self.my_proxy = tango.DeviceProxy("test/device1/1")
        self.set_state(tango.DevState.ON)

    @attribute
    def attr3(self):
        return self.my_proxy.attr1


def test_no_warnings_in_test_context():
    class TestDevice(Device): ...

    warnings.filterwarnings("error")
    with DeviceTestContext(TestDevice):
        pass


def test_single_device(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device1Asyncio):
            pass

    else:

        class TestDevice(Device1):
            green_mode = server_green_mode

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.attr1 == 100


def test_single_device_old_api():
    with DeviceTestContext(
        ClassicAPISimpleDeviceImpl, ClassicAPISimpleDeviceClass
    ) as proxy:
        assert proxy.attr1 == 100


def test_nested_single_device_in_same_process_failure():
    # Nesting not recommended - use MultiDeviceTestContext
    with DeviceTestContext(Device1, process=False):
        with pytest.raises(DevFailed):
            with DeviceTestContext(Device2, process=False):
                pass


def test_nested_single_device_in_different_processes_success_without_short_names():
    # Nesting not recommended - use MultiDeviceTestContext
    with DeviceTestContext(Device1, process=True) as proxy1:
        with DeviceTestContext(Device2, process=True) as proxy2:
            assert proxy1.attr1 == 100
            assert proxy2.attr2 == 200


def test_nested_single_device_in_different_processes_failure_with_short_names():
    # Nesting not recommended - use MultiDeviceTestContext
    with DeviceTestContext(Device1, device_name="test/nodb/1", process=True):
        with DeviceTestContext(Device2, process=True):
            with pytest.raises(DevFailed):
                DeviceProxy("test/nodb/1")


@pytest.mark.parametrize(
    "class_field, device",
    [
        (SimpleDevice, SimpleDevice),
        ("tango.test_utils.SimpleDevice", SimpleDevice),
        (
            (
                "tango.test_utils.ClassicAPISimpleDeviceClass",
                "tango.test_utils.ClassicAPISimpleDeviceImpl",
            ),
            ClassicAPISimpleDeviceImpl,
        ),
        (
            (
                "tango.test_utils.ClassicAPISimpleDeviceClass",
                ClassicAPISimpleDeviceImpl,
            ),
            ClassicAPISimpleDeviceImpl,
        ),
        (
            (
                ClassicAPISimpleDeviceClass,
                "tango.test_utils.ClassicAPISimpleDeviceImpl",
            ),
            ClassicAPISimpleDeviceImpl,
        ),
        (
            (ClassicAPISimpleDeviceClass, ClassicAPISimpleDeviceImpl),
            ClassicAPISimpleDeviceImpl,
        ),
    ],
)
def test_multi_devices_info(class_field, device):
    devices_info = ({"class": class_field, "devices": [{"name": "test/device1/1"}]},)

    dev_class = device if isinstance(device, str) else device.__name__

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        assert proxy1.info().dev_class == dev_class


def test_multi_with_single_device(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device1Asyncio):
            pass

    else:

        class TestDevice(Device1):
            green_mode = server_green_mode

    devices_info = ({"class": TestDevice, "devices": [{"name": "test/device1/1"}]},)

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        assert proxy1.attr1 == 100


def test_multi_with_single_device_old_api():
    devices_info = (
        {
            "class": (ClassicAPISimpleDeviceClass, ClassicAPISimpleDeviceImpl),
            "devices": [{"name": "test/device1/1"}],
        },
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        assert proxy1.attr1 == 100


def test_multi_with_two_devices(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class TestDevice1(Device1Asyncio):
            pass

        class TestDevice2(Device2Asyncio):
            pass

    else:

        class TestDevice1(Device1):
            green_mode = server_green_mode

        class TestDevice2(Device2):
            green_mode = server_green_mode

    devices_info = (
        {"class": TestDevice1, "devices": [{"name": "test/device1/1"}]},
        {"class": TestDevice2, "devices": [{"name": "test/device2/1"}]},
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        proxy2 = context.get_device("test/device2/1")
        assert proxy1.State() == tango.DevState.ON
        assert proxy2.State() == tango.DevState.ON
        assert proxy1.attr1 == 100
        assert proxy2.attr2 == 200


@pytest.mark.parametrize(
    "first_type, second_type, exception_type",
    [
        (Device1GreenModeUnspecified, Device2GreenModeUnspecified, None),
        (Device1GreenModeUnspecified, Device2Synchronous, None),
        (Device1GreenModeUnspecified, Device2Gevent, ValueError),
        (Device1GreenModeUnspecified, Device2Asyncio, ValueError),
        (Device1Synchronous, Device2GreenModeUnspecified, None),
        (Device1Synchronous, Device2Synchronous, None),
        (Device1Synchronous, Device2Gevent, ValueError),
        (Device1Synchronous, Device2Asyncio, ValueError),
        (Device1Asyncio, Device2GreenModeUnspecified, ValueError),
        (Device1Asyncio, Device2Synchronous, ValueError),
        (Device1Asyncio, Device2Gevent, ValueError),
        (Device1Asyncio, Device2Asyncio, None),
        (Device1Gevent, Device2GreenModeUnspecified, ValueError),
        (Device1Gevent, Device2Synchronous, ValueError),
        (Device1Gevent, Device2Gevent, None),
        (Device1Gevent, Device2Asyncio, ValueError),
    ],
)
def test_multi_with_mixed_device_green_modes(first_type, second_type, exception_type):
    devices_info = (
        {"class": first_type, "devices": [{"name": "test/device1/1"}]},
        {"class": second_type, "devices": [{"name": "test/device2/1"}]},
    )

    if exception_type is None:
        with MultiDeviceTestContext(devices_info):
            pass
    else:
        with pytest.raises(exception_type, match=r"mixed green mode"):
            with MultiDeviceTestContext(devices_info):
                pass


def stringify_green_mode(value):
    if isinstance(value, GreenMode):
        return str(value)
    else:
        return None


@pytest.mark.parametrize(
    "device_type, green_mode, global_mode, exception_type, executor_type",
    [
        # If a device specifies its green mode explicitly, then both
        # green_mode kwarg and global green mode are ignored. The device must use its specified mode.
        (
            Device1Synchronous,
            GreenMode.Asyncio,
            GreenMode.Asyncio,
            None,
            SynchronousExecutor,
        ),
        (
            Device1Synchronous,
            GreenMode.Gevent,
            GreenMode.Gevent,
            None,
            SynchronousExecutor,
        ),
        (
            Device1Asyncio,
            GreenMode.Synchronous,
            GreenMode.Synchronous,
            None,
            AsyncioExecutor,
        ),
        (Device1Asyncio, GreenMode.Gevent, GreenMode.Gevent, None, AsyncioExecutor),
        (
            Device1Gevent,
            GreenMode.Synchronous,
            GreenMode.Synchronous,
            None,
            GeventExecutor,
        ),
        (Device1Gevent, GreenMode.Asyncio, GreenMode.Asyncio, None, GeventExecutor),
        # If device doesn't specify its green mode, but green_mode kwarg is provided,
        # then we use green_mode kwarg
        (
            Device1GreenModeUnspecified,
            GreenMode.Synchronous,
            GreenMode.Asyncio,
            None,
            SynchronousExecutor,
        ),
        (
            Device1GreenModeUnspecified,
            GreenMode.Synchronous,
            GreenMode.Gevent,
            None,
            SynchronousExecutor,
        ),
        (
            Device1GreenModeUnspecified,
            GreenMode.Gevent,
            GreenMode.Synchronous,
            None,
            GeventExecutor,
        ),
        (
            Device1GreenModeUnspecified,
            GreenMode.Gevent,
            GreenMode.Asyncio,
            None,
            GeventExecutor,
        ),
        # Finally, if neither device green mode nor green_mode kwarg are specified, then use global mode instead.
        # (currently only works for synchronous mode - see unsupported modes below)
        (
            Device1GreenModeUnspecified,
            None,
            GreenMode.Synchronous,
            None,
            SynchronousExecutor,
        ),
        # Deprecated modes - starting from PyTango 10 it is deprecated
        # to modify sync servers to async "on the fly".
        # All base methods should be defined with "async def" instead.
        (
            Device1GreenModeUnspecified,
            GreenMode.Asyncio,
            GreenMode.Synchronous,
            DeprecationWarning,
            AsyncioExecutor,
        ),
        (
            Device1GreenModeUnspecified,
            GreenMode.Asyncio,
            GreenMode.Gevent,
            DeprecationWarning,
            AsyncioExecutor,
        ),
        # Unsupported modes - device servers with the following combinations
        # fail to start up. The cause is unknown. This could be fixed in the future.
        (
            Device1GreenModeUnspecified,
            None,
            GreenMode.Asyncio,
            RuntimeError,
            AsyncioExecutor,
        ),
        (
            Device1GreenModeUnspecified,
            None,
            GreenMode.Gevent,
            RuntimeError,
            GeventExecutor,
        ),
        (Device1Asyncio, None, GreenMode.Asyncio, RuntimeError, AsyncioExecutor),
        (Device1Gevent, None, GreenMode.Gevent, RuntimeError, GeventExecutor),
    ],
    ids=stringify_green_mode,
)
def test_green_modes_in_device_kwarg_and_global(
    device_type, green_mode, global_mode, exception_type, executor_type
):
    if WINDOWS and exception_type is not None:
        pytest.skip("Skip test that hangs on Windows")

    old_green_mode = tango.get_green_mode()
    try:
        tango.set_green_mode(global_mode)

        if exception_type is None:
            with DeviceTestContext(device_type, green_mode=green_mode):
                pass
        elif exception_type is DeprecationWarning:
            with pytest.warns((DeprecationWarning, RuntimeWarning)):
                with DeviceTestContext(device_type, green_mode=green_mode):
                    pass
        else:
            with pytest.raises(exception_type, match=r"stuck at init"):
                with DeviceTestContext(device_type, green_mode=green_mode, timeout=0.5):
                    pass
        assert type(get_worker()) is executor_type

    finally:
        tango.set_green_mode(old_green_mode)


def test_multi_with_async_devices_initialised():
    devices_info = (
        {"class": Device1Asyncio, "devices": [{"name": "test/device1/1"}]},
        {"class": Device2Asyncio, "devices": [{"name": "test/device2/1"}]},
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        proxy2 = context.get_device("test/device2/1")
        assert proxy1.State() == tango.DevState.ON
        assert proxy2.State() == tango.DevState.ON
        assert proxy1.attr1 == 100
        assert proxy2.attr2 == 200


def test_multi_device_access_via_test_context_methods():
    devices_info = (
        {"class": Device1, "devices": [{"name": "test/device1/1"}]},
        {"class": Device2, "devices": [{"name": "test/device2/2"}]},
    )

    with MultiDeviceTestContext(devices_info) as context:
        device_access1 = context.get_device_access("test/device1/1")
        device_access2 = context.get_device_access("test/device2/2")
        server_access = context.get_server_access()
        assert "test/device1/1" in device_access1
        assert "test/device2/2" in device_access2
        assert context.server_name in server_access
        proxy1 = tango.DeviceProxy(device_access1)
        proxy2 = tango.DeviceProxy(device_access2)
        proxy_server = tango.DeviceProxy(server_access)
        assert proxy1.attr1 == 100
        assert proxy2.attr2 == 200
        assert proxy_server.State() == tango.DevState.ON


def test_multi_short_name_device_proxy_access_without_tango_db():
    devices_info = ({"class": Device1, "devices": [{"name": "test/device1/1"}]},)

    with MultiDeviceTestContext(devices_info, process=True):
        proxy1 = tango.DeviceProxy("test/device1/1")
        assert proxy1.name() == "test/device1/1"
        assert proxy1.attr1 == 100


def test_multi_short_name_device_proxy_with_dependencies_access_without_tango_db():
    devices_info = (
        {"class": Device1, "devices": [{"name": "test/device1/1"}]},
        {"class": Device3, "devices": [{"name": "test/device3/3"}]},
    )

    with MultiDeviceTestContext(devices_info, process=True):
        proxy1 = tango.DeviceProxy("test/device1/1")
        proxy3 = tango.DeviceProxy("test/device3/3")
        assert proxy1.name() == "test/device1/1"
        assert proxy1.attr1 == 100
        assert proxy3.name() == "test/device3/3"
        assert proxy3.attr3 == 100
        proxy1.attr1 = 300
        assert proxy3.attr3 == 300


def test_multi_short_name_attribute_proxy_access_without_tango_db():
    devices_info = ({"class": Device1, "devices": [{"name": "test/device1/1"}]},)

    with MultiDeviceTestContext(devices_info):
        attr1 = tango.AttributeProxy("test/device1/1/attr1")
        assert attr1.name() == "attr1"
        assert attr1.read().value == 100


def test_single_short_name_device_proxy_access_without_tango_db():
    with DeviceTestContext(Device1, device_name="test/device1/1"):
        proxy1 = tango.DeviceProxy("test/device1/1")
        assert proxy1.name() == "test/device1/1"
        assert proxy1.attr1 == 100


def test_single_short_name_attribute_proxy_access_without_tango_db():
    with DeviceTestContext(Device1, device_name="test/device1/1"):
        attr1 = tango.AttributeProxy("test/device1/1/attr1")
        assert attr1.name() == "attr1"
        assert attr1.read().value == 100


def test_multi_short_name_access_fails_if_override_disabled():
    devices_info = ({"class": Device1, "devices": [{"name": "test/device1/a"}]},)

    context = MultiDeviceTestContext(devices_info, process=True)
    context.enable_test_context_tango_host_override = False
    context.start()
    try:
        with pytest.raises(DevFailed):
            _ = tango.DeviceProxy("test/device1/a")
    finally:
        context.stop()


def test_multi_device_proxy_cached():
    devices_info = ({"class": Device1, "devices": [{"name": "test/device1/1"}]},)

    with MultiDeviceTestContext(devices_info) as context:
        device1_first = context.get_device("test/device1/1")
        device1_second = context.get_device("test/device1/1")
        assert device1_first is device1_second


def test_multi_with_two_devices_with_properties(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class TestDevice1(Device):
            green_mode = server_green_mode

            prop1 = device_property(dtype=str)

            @command(dtype_out=str)
            async def get_prop1(self):
                return self.prop1

        class TestDevice2(Device):
            green_mode = server_green_mode

            prop2 = device_property(dtype=int)

            @command(dtype_out=int)
            async def get_prop2(self):
                return self.prop2

    else:

        class TestDevice1(Device):
            green_mode = server_green_mode

            prop1 = device_property(dtype=str)

            @command(dtype_out=str)
            def get_prop1(self):
                return self.prop1

        class TestDevice2(Device):
            green_mode = server_green_mode

            prop2 = device_property(dtype=int)

            @command(dtype_out=int)
            def get_prop2(self):
                return self.prop2

    devices_info = (
        {
            "class": TestDevice1,
            "devices": [{"name": "test/device1/1", "properties": {"prop1": "abcd"}}],
        },
        {
            "class": TestDevice2,
            "devices": [{"name": "test/device2/2", "properties": {"prop2": 5555}}],
        },
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        proxy2 = context.get_device("test/device2/2")
        assert proxy1.get_prop1() == "abcd"
        assert proxy2.get_prop2() == 5555


def test_multi_raises_on_invalid_file_database_properties():
    class TestDevice(Device):
        empty = device_property(dtype=(str,))

    with pytest.raises(RuntimeError, match="FileDatabase"):
        with DeviceTestContext(TestDevice, properties={"empty": []}):
            pass


@pytest.fixture(
    # Per test we have the input config tuple, and then the expected exception type
    params=[
        # empty config
        [tuple(), IndexError],
        # missing/invalid keys
        [({"not-class": Device1, "devices": [{"name": "test/device1/1"}]},), KeyError],
        [({"class": Device1, "not-devices": [{"name": "test/device1/1"}]},), KeyError],
        [({"class": Device1, "devices": [{"not-name": "test/device1/1"}]},), KeyError],
        # duplicate class
        [
            (
                {"class": Device1, "devices": [{"name": "test/device1/1"}]},
                {"class": Device1, "devices": [{"name": "test/device1/2"}]},
            ),
            ValueError,
        ],
        # mixing old "classic" API and new high level API
        [
            (
                {"class": Device1, "devices": [{"name": "test/device1/1"}]},
                {
                    "class": (ClassicAPISimpleDeviceClass, ClassicAPISimpleDeviceImpl),
                    "devices": [{"name": "test/device1/2"}],
                },
            ),
            ValueError,
        ],
        # mixing green modes
        [
            (
                {"class": Device1Synchronous, "devices": [{"name": "test/device1/1"}]},
                {"class": Device1Gevent, "devices": [{"name": "test/device1/2"}]},
            ),
            ValueError,
        ],
    ]
)
def bad_multi_device_config(request):
    return request.param


def test_multi_bad_config_fails(bad_multi_device_config):
    bad_config, expected_error = bad_multi_device_config
    with pytest.raises(expected_error):
        with MultiDeviceTestContext(bad_config):
            pass


@pytest.fixture()
def memorized_attribute_test_device_factory():
    """
    Returns a test device factory that provides a test device with an
    attribute that is memorized or not, according to its boolean
    argument
    """

    def _factory(is_attribute_memorized):
        class _Device(Device):
            def init_device(self):
                self._attr_value = 0

            attr = attribute(
                access=AttrWriteType.READ_WRITE,
                memorized=is_attribute_memorized,
                hw_memorized=is_attribute_memorized,
            )

            def read_attr(self):
                return self._attr_value

            def write_attr(self, value):
                self._attr_value = value

        return _Device

    return _factory


@pytest.mark.parametrize(
    "is_attribute_memorized, memorized_value, expected_value",
    [
        (False, None, 0),
        (False, "1", 0),
        (True, None, 0),
        (True, "1", 1),
    ],
)
def test_multi_with_memorized_attribute_values(
    memorized_attribute_test_device_factory,
    is_attribute_memorized,
    memorized_value,
    expected_value,
):
    TestDevice = memorized_attribute_test_device_factory(is_attribute_memorized)

    device_info = {"name": "test/device1/1"}
    if memorized_value is not None:
        device_info["memorized"] = {"attr": memorized_value}

    devices_info = ({"class": TestDevice, "devices": [device_info]},)

    with MultiDeviceTestContext(devices_info) as context:
        proxy = context.get_device("test/device1/1")
        assert proxy.attr == expected_value


@pytest.mark.parametrize(
    "is_attribute_memorized, memorized_value, expected_value",
    [
        (False, None, 0),
        (False, 1, 0),
        (True, None, 0),
        (True, 1, 1),
    ],
)
def test_single_with_memorized_attribute_values(
    memorized_attribute_test_device_factory,
    is_attribute_memorized,
    memorized_value,
    expected_value,
):
    TestDevice = memorized_attribute_test_device_factory(is_attribute_memorized)

    kwargs = (
        {"memorized": {"attr": memorized_value}} if memorized_value is not None else {}
    )

    with DeviceTestContext(TestDevice, **kwargs) as proxy:
        assert proxy.attr == expected_value


@pytest.mark.parametrize(
    "property_type, property_value, expected_outcome",
    [
        (str, {"prop": ""}, " "),
        ((str,), {"prop": ["", ""]}, (" ", " ")),
        ((str,), {"prop": np.array(["", ""])}, (" ", " ")),
    ],
)
def test_empty_string_property_bug(property_type, property_value, expected_outcome):
    class TestDevice(Device):
        prop = device_property(
            dtype=property_type,
        )

        @attribute(dtype=property_type, max_dim_x=10)
        def attr(self):
            return self.prop

    with DeviceTestContext(TestDevice, properties=property_value) as proxy:
        assert proxy.attr == expected_outcome


class AsyncDevice(Device):
    green_mode = GreenMode.Asyncio
    _value = 0
    _my_proxy = None

    @attribute
    async def value(self):
        return self._value

    @command(dtype_out="DevShort")
    async def increment(self):
        await asyncio.sleep(0.1)
        self._value += 1
        return self._value

    @attribute(dtype=str)
    async def my_name(self):
        return self.get_name()

    @attribute(dtype=str)
    async def child_name(self):
        attr = await self._my_proxy.read_attribute("my_name")
        return attr.value

    @command(dtype_in="DevString")
    async def connect_to_device(self, fqdn):
        proxy_future = get_device_proxy(fqdn, green_mode=GreenMode.Asyncio)
        assert isinstance(proxy_future, asyncio.Future)

        self._my_proxy = await proxy_future
        assert isinstance(self._my_proxy, DeviceProxy)


@pytest.mark.asyncio
@pytest.mark.parametrize("process", [True, False])
async def test_test_context_async_device_proxy(process):
    config = ({"class": AsyncDevice, "devices": [{"name": "test/device/1"}]},)

    with MultiDeviceTestContext(config, process=process) as context:
        fq_name = context.get_device_access("test/device/1")

        proxy_future = get_device_proxy(fq_name, green_mode=GreenMode.Asyncio)
        assert isinstance(proxy_future, asyncio.Future)

        proxy = await proxy_future
        assert isinstance(proxy, DeviceProxy)

        assert proxy.get_green_mode() == GreenMode.Asyncio

        # Ensure that we use the correct access method for the DeviceProxy call
        value = await proxy.increment()
        assert value == 1

        # Ensure that we use the correct access method for the DeviceProxy call
        attr = await proxy.read_attribute("value")
        assert attr.value == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("process", [True, False])
async def test_test_context_multi_async_device_proxy(process):
    config = (
        {
            "class": AsyncDevice,
            "devices": [{"name": "test/device/main"}, {"name": "test/device/child"}],
        },
    )

    with MultiDeviceTestContext(config, process=process) as context:
        fq_name_main = context.get_device_access("test/device/main")
        fq_name_child = context.get_device_access("test/device/child")

        proxy_future_main = get_device_proxy(fq_name_main, green_mode=GreenMode.Asyncio)
        assert isinstance(proxy_future_main, asyncio.Future)

        proxy_main = await proxy_future_main
        assert isinstance(proxy_main, DeviceProxy)

        assert proxy_main.get_green_mode() == GreenMode.Asyncio

        proxy_future_child = get_device_proxy(
            fq_name_child, green_mode=GreenMode.Asyncio
        )
        assert isinstance(proxy_future_child, asyncio.Future)

        proxy_child = await proxy_future_child
        assert isinstance(proxy_child, DeviceProxy)

        # Ensure that we use the correct access method for the DeviceProxy call
        await proxy_main.connect_to_device(fq_name_child)

        attr = await proxy_main.read_attribute("my_name")
        assert attr.value == "test/device/main"

        attr = await proxy_main.read_attribute("child_name")
        assert attr.value == "test/device/child"


class FutureAndGeventDevice(Device):
    green_mode = None
    _value = 0

    @attribute
    def value(self):
        return self._value

    @command(dtype_out="DevShort")
    def increment(self):
        self._value += 1
        return self._value


@pytest.mark.parametrize(
    "test_green_mode", [GreenMode.Futures, GreenMode.Gevent], ids=str
)
@pytest.mark.parametrize("process", [True, False])
def test_test_context_future_and_gevent_device_proxy(test_green_mode, process):
    FutureAndGeventDevice.green_mode = test_green_mode

    if test_green_mode == GreenMode.Futures:
        accessor = "result"
        expected_type = concurrent.futures.Future
    elif test_green_mode == GreenMode.Gevent:
        accessor = "get"
        expected_type = gevent.event.AsyncResult
    else:
        raise RuntimeError("Wrong GreenMode")

    config = ({"class": FutureAndGeventDevice, "devices": [{"name": "test/device/1"}]},)

    with MultiDeviceTestContext(config, process=process) as context:
        fq_name = context.get_device_access("test/device/1")

        proxy_future = get_device_proxy(fq_name, green_mode=test_green_mode, wait=False)
        assert isinstance(proxy_future, expected_type)

        proxy = getattr(proxy_future, accessor)()
        assert isinstance(proxy, DeviceProxy)

        assert proxy.get_green_mode() == test_green_mode

        # Ensure that we use the correct access method for the DeviceProxy call
        value_future = proxy.increment(wait=False)
        assert isinstance(value_future, expected_type)
        assert getattr(value_future, accessor)() == 1

        # Ensure that we use the correct access method for the DeviceProxy call
        attr_future = proxy.read_attribute("value", wait=False)
        assert isinstance(attr_future, expected_type)
        assert attr_future.result().value == 1


class FwdServer(Device):
    fwd_attr = attribute(forwarded=True)


@pytest.mark.parametrize("process", [True, False])
def test_forwarded_attributes(process):
    pytest.xfail(
        "Does not work yet - see https://gitlab.com/tango-controls/cppTango/-/issues/796"
    )
    devices_info = (
        {"class": Device1, "devices": [{"name": "test/device1/1"}]},
        {
            "class": FwdServer,
            "devices": [
                {
                    "name": "test/fwdserver/1",
                    "root_atts": {"fwd_attr": "test/device1/1/attr1"},
                }
            ],
        },
    )

    with MultiDeviceTestContext(devices_info, process=process):
        proxy_root = tango.DeviceProxy("test/device1/1")
        proxy_fwd = tango.DeviceProxy("test/fwdserver/1")
        assert "fwd_attr" in proxy_fwd.get_attribute_list()
        assert proxy_fwd.fwd_attr == 100
        proxy_fwd.fwd_attr = 200
        assert proxy_root.attr1 == 200
        proxy_root.attr1 = 300
        assert proxy_fwd.fwd_attr == 300


class DeviceStuckOnExit(Device):

    def delete_device(self):
        while True:
            time.sleep(1)


def test_device_context_thread_raises_if_device_server_stuck():
    context = DeviceTestContext(DeviceStuckOnExit)
    context.start()
    assert context.thread.daemon
    assert context.thread.is_alive()
    context.timeout = 0.2
    with pytest.raises(RuntimeError, match="failed to exit cleanly"):
        context.stop()
    # (thread can't be killed, so still alive here - it is a daemon so cleaned up
    # when test framework exits)


def test_device_context_process_raises_and_exits_if_device_server_stuck():
    context = DeviceTestContext(DeviceStuckOnExit, process=True)
    context.start()
    assert not context.thread.daemon  # don't want daemon for subprocess - kill instead
    assert context.thread.is_alive()
    context.timeout = 0.2
    with pytest.raises(RuntimeError, match=r"failed to exit cleanly.*kill subprocess"):
        context.stop()
    assert not context.thread.is_alive()


@pytest.mark.parametrize("process", [False, True])
def test_device_context_completes_if_device_server_not_stuck(process):
    context = DeviceTestContext(Device, process=process)
    context.start()
    assert context.thread.is_alive()
    context.stop()
    assert not context.thread.is_alive()



import pytest

from tango import (
    LockerLanguage,
    CmdArgType,
    MessBoxType,
    PollObjType,
    PollCmdCode,
    SerialModel,
    AttReqType,
    LockCmdCode,
    EventType,
    AttrSerialModel,
    KeepAliveCmdCode,
    AccessControlType,
    asyn_req_type,
    cb_sub_model,
    AttrQuality,
    AttrWriteType,
    AttrDataFormat,
    DevSource,
    ErrSeverity,
    DevState,
    DispLevel,
    AttrMemorizedType,
    ExtractAs,
    GreenMode,
    LogLevel,
    LogTarget,
    EventSubMode,
    EventReason,
)


@pytest.fixture(
    params=[
        LockerLanguage,
        CmdArgType,
        MessBoxType,
        PollObjType,
        PollCmdCode,
        SerialModel,
        AttReqType,
        LockCmdCode,
        EventType,
        AttrSerialModel,
        KeepAliveCmdCode,
        AccessControlType,
        asyn_req_type,
        cb_sub_model,
        AttrQuality,
        AttrWriteType,
        AttrDataFormat,
        DevSource,
        ErrSeverity,
        DevState,
        DispLevel,
        AttrMemorizedType,
        ExtractAs,
        GreenMode,
        LogLevel,
        LogTarget,
        EventSubMode,
        EventReason,
    ],
)
def enum_class(request):
    return request.param


def test_enum_str_and_int(enum_class):
    for name, value in list(enum_class.names.items()):
        assert str(value) == name
        assert int(value) == value


def test_enum_members_names_values(enum_class):
    members = enum_class.__members__
    members_as_ints = [int(enum_value) for enum_value in members.values()]
    names_to_enums = enum_class.names
    values_to_enums = enum_class.values

    assert names_to_enums == members
    assert list(values_to_enums.keys()) == members_as_ints
    assert list(values_to_enums.values()) == list(members.values())



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("init",)

__docformat__ = "restructuredtext"

import numpy

from tango.attribute_proxy import attribute_proxy_init
from tango.base_types import base_types_init
from tango.encoded_attribute import encoded_attribute_init
from tango.connection import connection_init
from tango.db import db_init
from tango.device_attribute import device_attribute_init
from tango.device_class import device_class_init
from tango.device_proxy import device_proxy_init
from tango.device_server import device_server_init
from tango.group import group_init
from tango.group_reply import group_reply_init
from tango.pytango_pprint import pytango_pprint_init
from tango.pyutil import pyutil_init
from tango.time_val import time_val_init
from tango.auto_monitor import auto_monitor_init
from tango._tango import constants
from tango._tango import _get_tango_lib_release

__INITIALIZED = False


def init_constants():
    import sys
    import platform

    tg_ver = tuple(map(int, constants.TgLibVers.split(".")))
    tg_ver_str = "0x%02d%02d%02d00" % (tg_ver[0], tg_ver[1], tg_ver[2])
    constants.TANGO_VERSION_HEX = int(tg_ver_str, 16)

    PYBIND11_VERSION = ".".join(
        map(
            str,
            (
                constants.PYBIND11_VERSION_MAJOR,
                constants.PYBIND11_VERSION_MINOR,
                constants.PYBIND11_VERSION_PATCH,
            ),
        )
    )
    constants.PYBIND11_VERSION = PYBIND11_VERSION

    class Compile:
        PY_VERSION = constants.PY_VERSION
        TANGO_VERSION = constants.TANGO_VERSION
        PYBIND11_VERSION = constants.PYBIND11_VERSION
        NUMPY_VERSION = constants.NUMPY_VERSION
        # UNAME = tuple(map(str, json.loads(constants.UNAME)))

    tg_rt_ver_nb = _get_tango_lib_release()
    tg_rt_major_ver = tg_rt_ver_nb // 100
    tg_rt_minor_ver = tg_rt_ver_nb // 10 % 10
    tg_rt_patch_ver = tg_rt_ver_nb % 10
    tg_rt_ver = ".".join(map(str, (tg_rt_major_ver, tg_rt_minor_ver, tg_rt_patch_ver)))

    class Runtime:
        PY_VERSION = ".".join(map(str, sys.version_info[:3]))
        TANGO_VERSION = tg_rt_ver
        NUMPY_VERSION = numpy.__version__
        UNAME = platform.uname()

    constants.Compile = Compile
    constants.Runtime = Runtime


def init():
    global __INITIALIZED
    if __INITIALIZED:
        return

    init_constants()
    base_types_init()
    encoded_attribute_init()
    connection_init()
    db_init()
    device_attribute_init()
    device_class_init()
    device_proxy_init()
    device_server_init()
    group_init()
    group_reply_init()
    pytango_pprint_init()
    pyutil_init()
    time_val_init()
    auto_monitor_init()

    # must come last: depends on device_proxy.init()
    attribute_proxy_init()

    __INITIALIZED = True



"""Basic check to see if stub file was generated and installed.

For use during continuous integration testing.
"""

import sys
from pathlib import Path

import tango


if __name__ == "__main__":
    pyi_path = Path(tango.__file__).parent / "_tango.pyi"
    if not pyi_path.is_file():
        print(f"Error: typing stub file is missing: {pyi_path}")
        sys.exit(1)
    else:
        print(f"Typing stub file exists: {pyi_path}")
